unit ApeCaches;

// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// "http://www.mozilla.org/MPL/"
// 
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
// 
// The Original Code is "ApeCaches.pas".
// 
// The Initial Developer of the Original Code is Dieter Khler (Heidelberg, 
// Germany, "http://www.philo.de/").
// Portions created by the Initial Developer are Copyright (C) 2002
// Dieter Khler. All Rights Reserved.
// 
// Alternatively, the contents of this file may be used under the terms of
// the GNU General Public License Version 2 or later (the "GPL"), in which 
// case the provisions of the GPL are applicable instead of those above. 
// If you wish to allow use of your version of this file only under the 
// terms of the GPL, and not to allow others to use your version of this 
// file under the terms of the MPL, indicate your decision by deleting the
// provisions above and replace them with the notice and other provisions
// required by the GPL. If you do not delete the provisions above, a
// recipient may use your version of this file under the terms of any one
// of the MPL or the GPL.

interface

uses
  StrUtils, SysUtils, Classes, XDOM_2_4;

resourcestring
  // Attention: string must not be longer then 255 characters
  MAPPINGLIST_TEMPLATE =
    '<!DOCTYPE ape:catalog [' +
    '<!ELEMENT ape:catalog (ape:source)*>' +
    '<!ELEMENT ape:source (ape:target)+>' +
    '<!ELEMENT ape:target (ape:title,ape:role+)>' +
    '<!ELEMENT ape:role (ape:title)>' +
    '<!ELEMENT ape:title (#PCDATA)>' +

    '<!ATTLIST ape:catalog ' +
    'ape:version CDATA #FIXED "1.0.0" ' +
    'xmlns:ape CDATA #FIXED "http://www.philo.de/ape/test-001" ' +
    'xml:base CDATA #IMPLIED>' +

    '<!ATTLIST ape:source ' +
    'ape:href CDATA #REQUIRED>' +

    '<!ATTLIST ape:target ' +
    'ape:href CDATA #REQUIRED>' +

    '<!ATTLIST ape:role ' +
    'ape:href CDATA #REQUIRED>' +

    ']>'+

    '<ape:catalog/>';

const
  xmlns_ape:   wideString  = 'http://www.philo.de/ape/test-001';
  xmlns_xlink: wideString  = 'http://www.w3.org/1999/xlink';

type

  TApeMappingCatalog = class;

  TApeFetchAction = (faAccept,faSkip);

  TApeBeforeEvaluateEvent = procedure(Sender: TObject; FileURL: wideString; var FetchAction: TApeFetchAction) of object;
  TApeBeginEvaluateEvent  = procedure(Sender: TObject; FileURL: wideString) of object;
  TApeEndEvaluateEvent    = procedure(Sender: TObject; FileURL: wideString; Success: boolean) of object;
  TApeInfoFoundEvent      = procedure(InfoFileURL, URN, URNLabel: wideString) of object;
  TApeMappingFoundEvent   = procedure(SourceURI, MappingFileURL: wideString; AObject: TObject) of object;

  TApeMappingCacheObject = class
  private
    FTargetURI: wideString;
    FTargetMappingURN: wideString;
  public
    property TargetMappingURN: wideString read FTargetMappingURN write FTargetMappingURN;
    property TargetURI: wideString read FTargetURI write FTargetURI;
  end;

  TApeMappingFetcher = class
  private
    FAfterEvaluateEvent: TApeEndEvaluateEvent;
    FBeforeEvaluateEvent: TApeBeforeEvaluateEvent;
    FDomImpl: TDomImplementation;
    FMappingCatalog: TApeMappingCatalog;
    FMappingFoundEvent: TApeMappingFoundEvent;
    FXmlDocBuilder: TXmlDocBuilder;
    FXmlStandardDomReader: TXmlStandardDomReader;
    FXmlToDomParser: TXmlToDomParser;
  protected
    procedure DoAfterEvaluate(FileURL: wideString;
                              Success: boolean); virtual;
    procedure DoBeforeEvaluate(FileURL: wideString;
                               var FetchAction: TApeFetchAction); virtual;
    procedure DoMappingFound(SourceURI,
                             MappingFileURL,
                             TargetURI,
                             TargetMappingURN: wideString); virtual;
  public
    constructor Create(AMappingCatalog: TApeMappingCatalog); virtual;
    destructor Destroy; override;
    function FetchMappings(const URN: wideString;
                           const UrlList: TdomNameValueList): boolean; virtual;
    property OnAfterEvaluate: TApeEndEvaluateEvent read FAfterEvaluateEvent write FAfterEvaluateEvent;
    property OnBeforeEvaluate: TApeBeforeEvaluateEvent read FBeforeEvaluateEvent write FBeforeEvaluateEvent;
    property OnMappingFound: TApeMappingFoundEvent read FMappingFoundEvent write FMappingFoundEvent;
  end;

  TApeMappingCache = class
  private
    FKeyList: TdomWideStringList;
    FSourceList: TdomWideStringList;
    FCacheSize: integer;
    procedure SetCacheSize(const Value: integer);
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Clear; virtual;
    function GetTargetAndMapping(const Key: WideString;
                                       R: TList): boolean; virtual;
    function SourceCached(Source: WideString): boolean; virtual;
    procedure Store(Key,
                    Source: WideString;
                    AObject: TObject); virtual;
    property CacheSize: integer read FCacheSize write SetCacheSize default -1;
  end;

  TApeMappingCacheManager = class
  private
    FBeginEvaluateEvent: TApeBeginEvaluateEvent;
    FEndEvaluateEvent: TApeEndEvaluateEvent;
    FMappingCache: TApeMappingCache;
    FMappingFetcher: TApeMappingFetcher;
    function GetCacheSize: integer;
    procedure SetCacheSize(const Value: integer);
  protected
    procedure AfterEvaluateMappingFetcher(Sender: TObject;
                                          FileURL: wideString;
                                          Success: boolean); virtual;
    procedure BeforeEvaluateMappingFetcher(Sender: TObject;
                                           FileURL: wideString;
                                           var FetchAction: TApeFetchAction); virtual;
    procedure DoBeginEvaluate(FileURL: wideString); virtual;
    procedure DoEndEvaluate(FileURL: wideString;
                            Success: boolean); virtual;
  public
    constructor Create(AMappingCatalog: TApeMappingCatalog); virtual;
    destructor Destroy; override;
    procedure Clear; virtual;
    function GetTargetAndMappingBySourceURI(const SourceURI: wideString;
                                            const UrlList: TdomNameValueList): boolean; virtual;
    property CacheSize: integer read GetCacheSize write SetCacheSize;
    property OnBeginEvaluate: TApeBeginEvaluateEvent read FBeginEvaluateEvent write FBeginEvaluateEvent;
    property OnEndEvaluate: TApeEndEvaluateEvent read FEndEvaluateEvent write FEndEvaluateEvent;
  end;

  TApeInfoFetcher = class
  private
    FAfterEvaluateEvent: TApeEndEvaluateEvent;
    FBeforeEvaluateEvent: TApeBeforeEvaluateEvent;
    FDomImpl: TDomImplementation;
    FInfoFoundEvent: TApeInfoFoundEvent;
    FMappingCatalog: TApeMappingCatalog;
    FXmlDocBuilder: TXmlDocBuilder;
    FXmlStandardDomReader: TXmlStandardDomReader;
    FXmlToDomParser: TXmlToDomParser;
  protected
    procedure DoAfterEvaluate(FileURL: wideString;
                              Success: boolean); virtual;
    procedure DoBeforeEvaluate(FileURL: wideString;
                               var FetchAction: TApeFetchAction); virtual;
    procedure DoInfoFound(InfoFileURL,
                          URN,
                          URNLabel: wideString); virtual;
  public
    constructor Create(AMappingCatalog: TApeMappingCatalog); virtual;
    destructor Destroy; override;
    function FetchInfos(const URN: wideString;
                          out URNLabel: wideString): boolean; virtual;
    property OnAfterEvaluate: TApeEndEvaluateEvent read FAfterEvaluateEvent write FAfterEvaluateEvent;
    property OnBeforeEvaluate: TApeBeforeEvaluateEvent read FBeforeEvaluateEvent write FBeforeEvaluateEvent;
    property OnInfoFound: TApeInfoFoundEvent read FInfoFoundEvent write FInfoFoundEvent;
  end;

  TApeInfoCache = class
  private
    FCachedFiles: TdomWideStringList;
    FCacheSize: integer;
    FUrnLabelList: TdomNameValueList;
    procedure SetCacheSize(const Value: integer);
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Add(InfoFileURL,
                  URN,
                  URNLabel: wideString); virtual;
    procedure Clear; virtual;
    function GetLabelByURN(const URN: wideString; out URNLabel: wideString): boolean; virtual;
    function IsFileCached(URL: wideString): boolean; virtual;
    property CacheSize: integer read FCacheSize write SetCacheSize default -1;
  end;

  TApeInfoCacheManager = class
  private
    FBeginEvaluateEvent: TApeBeginEvaluateEvent;
    FEndEvaluateEvent: TApeEndEvaluateEvent;
    FInfoCache: TApeInfoCache;
    FInfoFetcher: TApeInfoFetcher;
    function GetCacheSize: integer;
    procedure SetCacheSize(const Value: integer);
  protected
    procedure AfterEvaluateInfoFetcher(Sender: TObject;
                                       FileURL: wideString;
                                       Success: boolean); virtual;
    procedure BeforeEvaluateInfoFetcher(Sender: TObject;
                                        FileURL: wideString;
                                        var FetchAction: TApeFetchAction); virtual;
    procedure DoBeginEvaluate(FileURL: wideString); virtual;
    procedure DoEndEvaluate(FileURL: wideString; Success: boolean);
  public
    constructor Create(AMappingCatalog: TApeMappingCatalog); virtual;
    destructor Destroy; override;
    procedure Clear; virtual;
    function GetLabelByURN(const URN: wideString; out URNLabel: wideString): boolean; virtual;
    property CacheSize: integer read GetCacheSize write SetCacheSize;
    property OnBeginEvaluate: TApeBeginEvaluateEvent read FBeginEvaluateEvent write FBeginEvaluateEvent;
    property OnEndEvaluate: TApeEndEvaluateEvent read FEndEvaluateEvent write FEndEvaluateEvent;
  end;

  TApeMappingCatalog = class (TPersistent)
  private
    FSourceNameValueTree: TdomNameValueTree;
  protected
    procedure AssignTo(Dest: TPersistent); override;
    function BuildMappingCatalogDOM(const xmlToDomParser: TXmlToDomParser): TdomDocument; virtual;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Append(targetMappingURN,
                     sourceMappingURN,
                     targetURL,
                     targetTitle: wideString); virtual;
    procedure Assign(source: TPersistent); override;
    procedure Clear; virtual;
    procedure DeleteSourceUrn(index: integer); virtual;
    function GetAsUTF8Str: string; virtual;
    function GetAsUTF16BEStr: wideString; virtual;
    procedure RemoveByIndex(sourceIndex, targetIndex: integer); virtual;
    procedure RemoveByName(sourceURN, targetURL: wideString); virtual;
    function SourceMappingCount: integer; virtual;
    function SourceMappingURN(index: integer): wideString; virtual;
    function TargetCountByIndex(sourceIndex: integer): integer; virtual;
    function TargetCountByName(sourceURN: wideString): integer; virtual;
    function TargetMappingTitle(targetMappingURN: wideString): wideString; virtual;
    function TargetURLByIndex(sourceIndex, targetIndex: integer): wideString; virtual;
    function TargetURLByName(sourceURN: wideString; targetIndex: integer): wideString; virtual;
    function TargetTitleByIndex(sourceIndex, targetIndex: integer): wideString; virtual;
    function TargetTitleByName(sourceURN: wideString; targetIndex: integer): wideString; virtual;
    procedure UpdateFromDOM(Doc: TdomDocument); virtual;
  end;

implementation

{ TApeMappingFetcher }

constructor TApeMappingFetcher.Create(AMappingCatalog: TApeMappingCatalog);
begin
  FMappingCatalog:= AMappingCatalog;

  FDomImpl:= TDomImplementation.create(nil);
  FXmlDocBuilder:= TXmlDocBuilder.create(nil);
  FXmlStandardDomReader:= TXmlStandardDomReader.create(nil);
  FXmlToDomParser:= TXmlToDomParser.create(nil);
  FXmlToDomParser.DomImpl:= FDomImpl;
  with FXmlStandardDomReader do begin
    PrefixMapping:= true;
    SuppressXmlns:= true;
    NextHandler:= FXmlDocBuilder;
  end;
  FXmlDocBuilder.BuildNamespaceTree:= true;
end;

destructor TApeMappingFetcher.Destroy;
begin
  FDomImpl.free;
  FXmlDocBuilder.free;
  FXmlStandardDomReader.free;
  FXmlToDomParser.free;

  inherited;
end;

procedure TApeMappingFetcher.DoAfterEvaluate(FileURL: wideString;
                                             Success: boolean);
begin
  if assigned(FAfterEvaluateEvent)
    then FAfterEvaluateEvent(self, FileURL, Success);
end;

procedure TApeMappingFetcher.DoBeforeEvaluate(FileURL: wideString;
  var FetchAction: TApeFetchAction);
begin
  FetchAction:= faAccept;
  if assigned(FBeforeEvaluateEvent)
    then FBeforeEvaluateEvent(self, FileURL, FetchAction);
end;

procedure TApeMappingFetcher.DoMappingFound(SourceURI,
                                            MappingFileURL,
                                            TargetURI,
                                            TargetMappingURN: wideString);
var
  MappingCacheObject: TApeMappingCacheObject;
begin
  if assigned(FMappingFoundEvent) then begin
    MappingCacheObject:= TApeMappingCacheObject.Create;
    MappingCacheObject.TargetURI:= TargetURI;
    MappingCacheObject.TargetMappingURN:= TargetMappingURN;
    FMappingFoundEvent(SourceURI, MappingFileURL, MappingCacheObject);
  end;
end;

function TApeMappingFetcher.FetchMappings(const URN: wideString;
                                          const UrlList: TdomNameValueList): boolean;
var
  mappingFileURL,targetMappingURN,sourceURI,targetURI,urnValue: wideString;
  i,j: integer;
  dummyDoc, mappingDoc: TdomDocument;
  ape_body,ape_Seq,ape_item: TdomElement;
  xlink_href: TdomAttr;
  path: TFilename;
  dummy1,dummy2,dummy3: string;
  ok: boolean;
  FetchAction: TApeFetchAction;
begin
  result:= false;
  if (URN = '') then exit;
  if not assigned(FMappingCatalog) then exit;

//  if not FUnsuccesfulAnlyzedURNs.find(URN,dummyIndex) then begin

  with FMappingCatalog do begin
    for i:= 0 to pred(SourceMappingCount) do begin
      urnValue:= SourceMappingURN(i);
      if urnValue = LeftStr(URN,system.length(urnValue)) then begin // xxx is this too simple ?
        for j:= 0 to pred(TargetCountByName(urnValue)) do begin     // xxx index instead of urnValue might be faster
          mappingFileURL:= TargetURLByName(urnValue,j);

          DoBeforeEvaluate(mappingFileURL,FetchAction);
          if FetchAction = faAccept then begin

            ok:= UriWideStrToFilename(mappingFileURL,path,dummy1,dummy2,dummy3);
            if ok then begin
              ok:= fileExists(path);
              if ok then begin

                FDOMImpl.clear;
                with FXmlToDomParser do begin
                  try
                    dummyDoc:= FileToDom(Path);
                    dummyDoc.prepareASModels;
                    if ok
                      then ok:= dummyDoc.validate(erReplace);
                    if ok then begin

                      mappingDoc:= FDOMImpl.createDocument('dummy',nil);
                      mappingDoc.clear;
                      FXmlDocBuilder.referenceNode:= mappingDoc;
                      ok:= FXmlStandardDomReader.parse(dummyDoc);
                      if ok then begin

                        mappingDoc.setBaseUri(FilenameToUriStr(path,[fuPlainColon]));
                        with mappingDoc.documentElement do begin
                          targetMappingURN:= getAttributeNSNormalizedValue(xmlns_ape,'target');
                          ape_body:= getFirstChildElementNS(xmlns_ape,'body');
                          if assigned(ape_body) then begin
                            ape_Seq:= ape_body.getFirstChildElementNS(xmlns_ape,'seq');
                            while assigned(ape_Seq) do begin
                              sourceURI:= ape_Seq.getAttributeNSNormalizedValue(xmlns_ape,'about');
                              if sourceURI <> '' then begin
                                ape_item:= ape_Seq.getFirstChildElementNS(xmlns_ape,'item');
                                while assigned(ape_item) do begin
                                  xlink_href:= ape_item.getAttributeNodeNS(xmlns_xlink,'href');
                                  if assigned(xlink_href) then begin
                                    resolveRelativeUriWideStr(xlink_href.baseUri,xlink_href.value,targetUri);
                                    if targetUri <> '' then begin

                                      DoMappingFound(sourceURI,mappingFileURL,targetURI,targetMappingURN);

                                      if sourceURI = URN then begin
                                        UrlList.add(targetURI,targetMappingURN);
                                        result:= true;
                                      end;

                                     end;
                                  end;
                                  ape_item:= ape_item.getNextSiblingElementNS(xmlns_ape,'item');
                                end; {while ...}
                              end; {if ...}
                              ape_Seq:= ape_Seq.getNextSiblingElementNS(xmlns_ape,'seq');
                            end; {while ...}
                          end; {if ...}
                        end; {with mappingDoc.documentElement ...}

                      end; {if ...}
                    end; {if ...}
                  except
                    result:= false;
                  end;
                end; {with FXmlToDomParser ...}

              end; {if ...}
            end; {if ...}

            DoAfterEvaluate(mappingFileURL,ok);
          end; {if ...}

        end; {for ...}
      end; {if ...}
    end; {for ...}
  end; {with ...}
end;

{ TApeMappingCache }

constructor TApeMappingCache.Create;
begin
  FCacheSize:= -1;

  FKeyList:= TdomWideStringList.Create;
  with FKeyList do begin
    sorted:= true;
    duplicates:= dupError;
  end;
  FSourceList:= TdomWideStringList.Create;
  with FSourceList do begin
    sorted:= true;
    duplicates:= dupIgnore;
  end;
end;

destructor TApeMappingCache.Destroy;
begin
  Clear;
  FKeyList.free;
  FSourceList.free;
  inherited;
end;

procedure TApeMappingCache.Clear;
var
  index1, index2: integer;
begin
  with FKeyList do begin
    for index1:= 0 to pred(count) do
      TdomWideStringList(objects[index1]).free;
    clear;
  end;
  with FSourceList do begin
    for index1:= 0 to pred(count) do begin
      with TdomWideStringList(objects[index1]) do
        for index2:= 0 to pred(count) do
          objects[index2].free;
      TdomWideStringList(objects[index1]).free;
    end;
    clear;
  end;
end;

function TApeMappingCache.GetTargetAndMapping(
  const Key: WideString; R: TList): boolean;
var
  I, Index: integer;
begin
  with FKeyList do begin
    if (Count > FCacheSize) and (FCacheSize > -1) then begin
      Self.Clear;
      Result:= false;
    end else begin
      Result:= Find(Key,Index);
      if Result then
        with TdomWideStringList(Objects[Index]) do
          for I:= 0 to Pred(Count) do
            R.Add(Objects[I]);
    end;
  end;
end;

function TApeMappingCache.SourceCached(Source: WideString): boolean;
var
  Index: Integer;
begin
  Result:= FSourceList.Find(Source,Index);
end;

procedure TApeMappingCache.SetCacheSize(const Value: integer);
begin
  FCacheSize := Value;
end;

procedure TApeMappingCache.Store(Key, Source: WideString; AObject: TObject);
var
  SourceObjectList: TdomWideStringList;
  index: integer;
begin
  with FKeyList do begin
    if find(Key,index) then begin
      SourceObjectList:= TdomWideStringList(objects[index]);
    end else begin
      SourceObjectList:= TdomWideStringList.Create;
      addObject(Key,SourceObjectList);
    end;
  end;
  SourceObjectList.addObject(Source,AObject);  // Note: Duplicates are allowed.

  with FSourceList do begin
    if find(Source,index) then begin
      SourceObjectList:= TdomWideStringList(objects[index]);
    end else begin
      SourceObjectList:= TdomWideStringList.Create;
      addObject(Source,SourceObjectList);
    end;
  end;
  SourceObjectList.addObject(Key,AObject);  // Note: Duplicates are allowed.
end;


{ TApeMappingCacheManager }

constructor TApeMappingCacheManager.Create(AMappingCatalog: TApeMappingCatalog);
begin
  FMappingCache:= TApeMappingCache.Create;
  FMappingFetcher:= TApeMappingFetcher.Create(AMappingCatalog);
  with FMappingFetcher do begin
    OnAfterEvaluate:= AfterEvaluateMappingFetcher;
    OnBeforeEvaluate:= BeforeEvaluateMappingFetcher;
    OnMappingFound:= FMappingCache.Store;
  end;
end;

destructor TApeMappingCacheManager.Destroy;
begin
  FMappingCache.free;
  FMappingFetcher.free;
  inherited;
end;

procedure TApeMappingCacheManager.AfterEvaluateMappingFetcher(
  Sender: TObject; FileURL: wideString; Success: boolean);
begin
  DoEndEvaluate(FileURL,Success);
end;

procedure TApeMappingCacheManager.BeforeEvaluateMappingFetcher(
  Sender: TObject; FileURL: wideString; var FetchAction: TApeFetchAction);
begin
  if FMappingCache.SourceCached(FileURL)
    then FetchAction:= faSkip
    else DoBeginEvaluate(FileURL);
end;

procedure TApeMappingCacheManager.Clear;
begin
  FMappingCache.Clear;
end;

procedure TApeMappingCacheManager.DoBeginEvaluate(FileURL: wideString);
begin
  if assigned(FBeginEvaluateEvent)
    then FBeginEvaluateEvent(self, FileURL);
end;

procedure TApeMappingCacheManager.DoEndEvaluate(FileURL: wideString;
  Success: boolean);
begin
  if assigned(FEndEvaluateEvent)
    then FEndEvaluateEvent(self, FileURL, Success);
end;

function TApeMappingCacheManager.GetCacheSize: integer;
begin
  Result:= FMappingCache.CacheSize;
end;

function TApeMappingCacheManager.GetTargetAndMappingBySourceURI(
  const SourceURI: wideString; const UrlList: TdomNameValueList): boolean;
var
  List: TList;
  I: Integer;
begin
  List:= TList.Create;
  try
    Result:= FMappingCache.GetTargetAndMapping(SourceURI,List);
    UrlList.clear;
    for I:= 0 to Pred(List.Count) do
      UrlList.add(TApeMappingCacheObject(List[I]).TargetURI,
                  TApeMappingCacheObject(List[I]).TargetMappingURN);
  finally
    List.free;
  end;
  if not result
    then result:= FMappingFetcher.FetchMappings(SourceURI,UrlList);
end;

procedure TApeMappingCacheManager.SetCacheSize(const Value: integer);
begin
  FMappingCache.CacheSize:= Value;
end;

{ TApeInfoFetcher }

constructor TApeInfoFetcher.Create(AMappingCatalog: TApeMappingCatalog);
begin
  FMappingCatalog:= AMappingCatalog;

  FDomImpl:= TDomImplementation.create(nil);
  FXmlDocBuilder:= TXmlDocBuilder.create(nil);
  FXmlStandardDomReader:= TXmlStandardDomReader.create(nil);
  FXmlToDomParser:= TXmlToDomParser.create(nil);
  FXmlToDomParser.DomImpl:= FDomImpl;
  with FXmlStandardDomReader do begin
    PrefixMapping:= true;
    SuppressXmlns:= true;
    NextHandler:= FXmlDocBuilder;
  end;
  FXmlDocBuilder.BuildNamespaceTree:= true;
end;

destructor TApeInfoFetcher.Destroy;
begin
  FDomImpl.free;
  FXmlDocBuilder.free;
  FXmlStandardDomReader.free;
  FXmlToDomParser.free;
  inherited;
end;

procedure TApeInfoFetcher.DoAfterEvaluate(FileURL: wideString;
                                          Success: boolean);
begin
  if assigned(FAfterEvaluateEvent)
    then FAfterEvaluateEvent(self, FileURL, Success);
end;

procedure TApeInfoFetcher.DoBeforeEvaluate(FileURL: wideString;
  var FetchAction: TApeFetchAction);
begin
  FetchAction:= faAccept;
  if assigned(FBeforeEvaluateEvent)
    then FBeforeEvaluateEvent(self, FileURL, FetchAction);
end;

procedure TApeInfoFetcher.DoInfoFound(InfoFileURL, URN,
  URNLabel: wideString);
begin
  if assigned(FInfoFoundEvent)
    then FInfoFoundEvent(InfoFileURL, URN, URNLabel);
end;

function TApeInfoFetcher.FetchInfos(const URN: wideString;
  out URNLabel: wideString): boolean;
var
  infoFileURL,aboutURN,itemLabel,urnValue: wideString;
  i,j: integer;
  dummyDoc, mappingDoc: TdomDocument;
  ape_body,ape_infoItem: TdomElement;
  path: TFilename;
  dummy1,dummy2,dummy3: string;
  ok: boolean;
  FetchAction: TApeFetchAction;
begin
  result:= false;
  if (URN = '') then exit;
  if not assigned(FMappingCatalog) then exit;

//  if not FUnsuccesfulAnlyzedURNs.find(URN,dummyIndex) then begin      // xxx ????

  with FMappingCatalog do begin
    for i:= 0 to pred(SourceMappingCount) do begin
      urnValue:= SourceMappingURN(i);
      if urnValue = LeftStr(URN,system.length(urnValue)) then begin // xxx is this too simple ?
        for j:= 0 to pred(TargetCountByName(urnValue)) do begin     // xxx index instead of urnValue might be faster
          infoFileURL:= TargetURLByName(urnValue,j);

          DoBeforeEvaluate(infoFileURL,FetchAction);
          if FetchAction = faAccept then begin

            ok:= UriWideStrToFilename(infoFileURL,path,dummy1,dummy2,dummy3);
            if ok then begin
              ok:= fileExists(path);
              if ok then begin

                FDOMImpl.clear;
                with FXmlToDomParser do begin
                  try
                    dummyDoc:= FileToDom(Path);
                    dummyDoc.prepareASModels;
                    if ok
                      then ok:= dummyDoc.validate(erReplace);
                    if ok then begin

                      mappingDoc:= FDOMImpl.createDocument('dummy',nil);
                      mappingDoc.clear;
                      FXmlDocBuilder.referenceNode:= mappingDoc;
                      ok:= FXmlStandardDomReader.parse(dummyDoc);
                      if ok then begin

                        with mappingDoc.documentElement do begin
                          ape_body:= getFirstChildElementNS(xmlns_ape,'body');
                          if assigned(ape_body) then begin
                            ape_infoItem:= ape_body.getFirstChildElementNS(xmlns_ape,'infoItem');
                            while assigned(ape_infoItem) do begin
                              aboutURN:= ape_infoItem.getAttributeNSNormalizedValue(xmlns_ape,'about');
                              if aboutURN <> '' then begin
                                itemLabel:= ape_infoItem.getAttributeNSNormalizedValue(xmlns_ape,'label');

                                DoInfoFound(infoFileURL,aboutURN,itemLabel);

                                if aboutURN = URN then begin
                                  URNLabel:= itemLabel;
                                  result:= true;
                                end;

                              end; {if ...}
                              ape_infoItem:= ape_infoItem.getNextSiblingElementNS(xmlns_ape,'infoItem');
                            end; {while ...}
                          end; {if ...}
                        end; {with mappingDoc.documentElement ...}

                      end; {if ...}
                    end; {if ...}
                  except
                    result:= false;
                  end;
                end; {with FXmlToDomParser ...}

              end; {if ...}
            end; {if ...}

            DoAfterEvaluate(infoFileURL,ok);
          end; {if ...}

        end; {for ...}
      end; {if ...}
    end; {for ...}
  end; {with ...}
end;

{ TApeInfoCache }

constructor TApeInfoCache.Create;
begin
  FUrnLabelList:= TdomNameValueList.create;
  FCachedFiles:= TdomWideStringList.create;
end;

destructor TApeInfoCache.Destroy;
begin
  FUrnLabelList.free;
  FCachedFiles.free;
  inherited;
end;

procedure TApeInfoCache.Add(InfoFileURL, URN, URNLabel: wideString);
begin
  FCachedFiles.Add(InfoFileURL);
  FUrnLabelList.Add(URN,URNLabel);
end;

procedure TApeInfoCache.Clear;
begin
  FUrnLabelList.clear;
  FCachedFiles.clear;
end;

function TApeInfoCache.GetLabelByURN(const URN: wideString; out URNLabel: wideString): boolean;
var
  index: integer;
begin
  with FUrnLabelList do begin
    if (Length > FCacheSize) and (FCacheSize > -1) then begin
      Self.Clear;
      Result:= false;
    end else begin
      index:= indexOfName(URN);
      result:= (index > -1);
      if result
        then URNLabel:= values[index]
        else URNLabel:= '';
    end;
  end;
end;

function TApeInfoCache.IsFileCached(URL: wideString): boolean;
var
  index: integer;
begin
  result:= FCachedFiles.find(URL,index);
end;

procedure TApeInfoCache.SetCacheSize(const Value: integer);
begin
  FCacheSize := Value;
end;

{ TApeInfoCacheManager }

constructor TApeInfoCacheManager.Create(AMappingCatalog: TApeMappingCatalog);
begin
  FInfoCache:= TApeInfoCache.Create;
  FInfoFetcher:= TApeInfoFetcher.Create(AMappingCatalog);
  with FInfoFetcher do begin
    OnAfterEvaluate:= AfterEvaluateInfoFetcher;
    OnBeforeEvaluate:= BeforeEvaluateInfoFetcher;
    OnInfoFound:= FInfoCache.Add;
  end;
end;

destructor TApeInfoCacheManager.Destroy;
begin
  FInfoCache.free;
  FInfoFetcher.free;
  inherited;
end;

procedure TApeInfoCacheManager.AfterEvaluateInfoFetcher(
  Sender: TObject; FileURL: wideString; Success: boolean);
begin
  DoEndEvaluate(FileURL,Success);
end;

procedure TApeInfoCacheManager.BeforeEvaluateInfoFetcher(
  Sender: TObject; FileURL: wideString; var FetchAction: TApeFetchAction);
begin
  if FInfoCache.IsFileCached(FileURL)
    then FetchAction:= faSkip
    else DoBeginEvaluate(FileURL);
end;

procedure TApeInfoCacheManager.Clear;
begin
  FInfoCache.clear;
end;

procedure TApeInfoCacheManager.DoBeginEvaluate(FileURL: wideString);
begin
  if assigned(FBeginEvaluateEvent)
    then FBeginEvaluateEvent(self, FileURL);
end;

procedure TApeInfoCacheManager.DoEndEvaluate(FileURL: wideString;
  Success: boolean);
begin
  if assigned(FEndEvaluateEvent)
    then FEndEvaluateEvent(self, FileURL, Success);
end;

function TApeInfoCacheManager.GetCacheSize: integer;
begin
  Result:= FInfoCache.CacheSize;
end;

function TApeInfoCacheManager.GetLabelByURN(const URN: wideString; out URNLabel: wideString): boolean;
begin
  result:= FInfoCache.GetLabelByURN(URN,URNLabel);
  if not result
    then result:= FInfoFetcher.FetchInfos(URN,URNLabel);
end;

procedure TApeInfoCacheManager.SetCacheSize(const Value: integer);
begin
  FInfoCache.CacheSize:= Value;
end;

{ TApeMappingCatalog }

constructor TApeMappingCatalog.Create;
begin
  FSourceNameValueTree:= TdomNameValueTree.Create;
  with FSourceNameValueTree do begin
    sorted:= true;
    duplicates:= dupError;
  end;
end;

destructor TApeMappingCatalog.Destroy;
begin
  FSourceNameValueTree.free;
  inherited;
end;

procedure TApeMappingCatalog.Append(targetMappingURN,
                                    sourceMappingURN,
                                    targetURL,
                                    targetTitle: wideString);
var
  newTargetMappingNameValueTree: TdomNameValueTree;
  newTargetNameValueTree: TdomNameValueTree;
  sourceIndex: integer;
  targetIndex: integer;
begin
  if (sourceMappingURN = '') or (targetURL = '') then exit;
  with FSourceNameValueTree do begin
    if not findOfName(sourceMappingURN,sourceIndex) then begin
      newTargetNameValueTree:= TdomNameValueTree.create;
      with newTargetNameValueTree do begin
        sorted:= true;
        duplicates:= dupError;
      end;
      sourceIndex:= addChild(sourceMappingURN,'',newTargetNameValueTree);
    end; {if ...}

    with children[sourceIndex] do begin // = TargetNameValueTree

      if not findOfName(targetURL,targetIndex) then begin
        newTargetMappingNameValueTree:= TdomNameValueTree.create;
        with newTargetMappingNameValueTree do begin
          sorted:= true;
          duplicates:= dupError;
        end;
        targetIndex:= addChild(targetURL,targetTitle,newTargetMappingNameValueTree);
      end; {if ...}

      children[targetIndex].add(targetMappingURN,targetTitle);  // xxx test for duplicate targetMappingURNs ?????

    end; {with ...}
  end; {with ...}
end;


procedure TApeMappingCatalog.Assign(source: TPersistent);
begin
  if (source is TApeMappingCatalog) or (source is TdomNameValueList) then begin
    if source = self then exit;
    FSourceNameValueTree.assign(source);
    exit;
  end;
  inherited assign(Source);
end;

procedure TApeMappingCatalog.AssignTo(Dest: TPersistent);
begin
  if dest is TdomNameValueList then begin
    TdomNameValueList(dest).assign(FSourceNameValueTree);
    exit;
  end;
  inherited assignTo(dest);
end;

function TApeMappingCatalog.BuildMappingCatalogDOM(const xmlToDomParser: TXmlToDomParser): TdomDocument;
var
  ape_role: TdomElement;
  ape_source: TdomElement;
  ape_target: TdomElement;
  ape_title: TdomElement;
  i: integer;
  j: integer;
  k: integer;
  titleText: TdomText;
begin
  result:= xmlToDomParser.stringToDom(MAPPINGLIST_TEMPLATE);
  for i:= 0 to pred(FSourceNameValueTree.length) do begin
    ape_source:= result.createElement('ape:source');
    result.documentElement.appendChild(ape_source);
    ape_source.setAttribute('ape:href',FSourceNameValueTree.names[i]);
    for j:= 0 to pred(FSourceNameValueTree.children[i].length) do begin
      ape_target:= result.createElement('ape:target');
      ape_source.appendChild(ape_target);
      ape_target.setAttribute('ape:href',FSourceNameValueTree.children[i].names[j]);
      ape_title:= result.createElement('ape:title');
      ape_target.appendChild(ape_title);
      titleText:= result.createTextNode(FSourceNameValueTree.children[i].values[j]);
      ape_title.appendChild(titleText);
      for k:= 0 to pred(FSourceNameValueTree.children[i].children[j].length) do begin
        ape_role:= result.createElement('ape:role');
        ape_target.appendChild(ape_role);
        ape_role.setAttribute('ape:href',FSourceNameValueTree.children[i].children[j].names[k]);
        ape_title:= result.createElement('ape:title');
        ape_role.appendChild(ape_title);
        titleText:= result.createTextNode(FSourceNameValueTree.children[i].children[j].values[k]);
        ape_title.appendChild(titleText);
      end;
    end;
  end;
end;

procedure TApeMappingCatalog.Clear;
begin
  FSourceNameValueTree.clear;
end;

procedure TApeMappingCatalog.DeleteSourceUrn(index: integer);
begin
  FSourceNameValueTree.Delete(index);
end;

function TApeMappingCatalog.GetAsUTF16BEStr: wideString;
var
  mappingCatalogDoc: TDomDocument;
  domImpl: TDomImplementation;
  xmlToDomParser: TXmlToDomParser;
  domToXmlParser: TDomToXmlParser;
begin
  domImpl:= TDomImplementation.create(nil);
  xmlToDomParser:= TXmlToDomParser.create(nil);
  xmlToDomParser.DomImpl:= domImpl;
  domToXmlParser:= TDomToXmlParser.create(nil);
  domToXmlParser.DomImpl:= domImpl;
  try
    mappingCatalogDoc:= BuildMappingCatalogDOM(xmlToDomParser);
    domToXmlParser.writeToWideString(mappingCatalogDoc,result);
  finally
    domToXmlParser.free;
    xmlToDomParser.free;
    domImpl.free;
  end;
end;

function TApeMappingCatalog.GetAsUTF8Str: string;
var
  mappingCatalogDoc: TDomDocument;
  domImpl: TDomImplementation;
  xmlToDomParser: TXmlToDomParser;
  domToXmlParser: TDomToXmlParser;
begin
  result:= '';
  domImpl:= TDomImplementation.create(nil);
  xmlToDomParser:= TXmlToDomParser.create(nil);
  xmlToDomParser.DomImpl:= domImpl;
  domToXmlParser:= TDomToXmlParser.create(nil);
  domToXmlParser.DomImpl:= domImpl;
  try
    mappingCatalogDoc:= BuildMappingCatalogDOM(xmlToDomParser);
    domToXmlParser.writeToString(mappingCatalogDoc,'UTF-8',result);
  finally
    domToXmlParser.free;
    xmlToDomParser.free;
    domImpl.free;
  end;
end;

procedure TApeMappingCatalog.RemoveByIndex(sourceIndex,
  targetIndex: integer);
var
  targetNameValueTree: TdomNameValueTree;
begin
  if (sourceIndex < 0) or (targetIndex < 0) then exit;

  // Delete Target:
  targetNameValueTree:= FSourceNameValueTree.children[sourceIndex];
  targetNameValueTree.Delete(targetIndex);

  // If source is empty then delete source, too:
  if targetNameValueTree.length = 0
    then DeleteSourceUrn(sourceIndex);
end;

procedure TApeMappingCatalog.RemoveByName(sourceURN,
  targetURL: wideString);
var
  sourceIndex,targetIndex: integer;
  targetNameValueTree: TdomNameValueTree;
begin
  with FSourceNameValueTree do begin
    if not findOfName(sourceURN,sourceIndex) then exit;           // Source name found?
    targetNameValueTree:= children[sourceIndex];
  end;
  if targetNameValueTree.findOfName(targetURL,targetIndex) then begin  // Target name found?
    targetNameValueTree.Delete(targetIndex); // Delete Target.
    if targetNameValueTree.length = 0        // If source is empty ...
      then DeleteSourceUrn(sourceIndex);     // ... then delete source, too.
  end;
end;

function TApeMappingCatalog.SourceMappingCount: integer;
begin
  result:= FSourceNameValueTree.length;
end;

function TApeMappingCatalog.SourceMappingURN(index: integer): wideString;
begin
  result:= FSourceNameValueTree.names[index];
end;

function TApeMappingCatalog.TargetCountByIndex(
  sourceIndex: integer): integer;
begin
  result:= FSourceNameValueTree.children[sourceIndex].length;
end;

function TApeMappingCatalog.TargetCountByName(
  sourceURN: wideString): integer;
var
  sourceIndex: integer;
begin
  if FSourceNameValueTree.findOfName(sourceURN,sourceIndex)
    then result:= TargetCountByIndex(sourceIndex)
    else result:= -1;
end;

function TApeMappingCatalog.TargetMappingTitle(targetMappingURN: wideString): wideString;
// xxx Very slow for large Catalogs.
var
  i: integer;
  j: integer;
  k: integer;
begin
  with FSourceNameValueTree do
    for i:= 0 to pred(length) do
      with children[i] do
        for j:= 0 to pred(length) do
          with children[j] do
            for k:= 0 to pred(length) do
              if names[k] = targetMappingURN then begin
                result:= values[k];
                exit;
              end;
end;

function TApeMappingCatalog.TargetTitleByIndex(sourceIndex,
  targetIndex: integer): wideString;
begin
  if sourceIndex < 0 then
    raise EStringListError.CreateFmt('List index out of bounds (%d)', [sourceIndex]);
  result:= FSourceNameValueTree.children[sourceIndex].values[targetIndex];
end;

function TApeMappingCatalog.TargetTitleByName(sourceURN: wideString;
  targetIndex: integer): wideString;
var
  sourceIndex: integer;
begin
  if FSourceNameValueTree.findOfName(sourceURN,sourceIndex)
    then result:= TargetTitleByIndex(sourceIndex,targetIndex)
    else result:= '';
end;

function TApeMappingCatalog.TargetURLByIndex(sourceIndex,
  targetIndex: integer): wideString;
begin
  if sourceIndex < 0 then
    raise EStringListError.CreateFmt('List index out of bounds (%d)', [sourceIndex]);
  result:= FSourceNameValueTree.children[sourceIndex].names[targetIndex];
end;

function TApeMappingCatalog.TargetURLByName(sourceURN: wideString;
  targetIndex: integer): wideString;
var
  sourceIndex: integer;
begin
  if FSourceNameValueTree.findOfName(sourceURN,sourceIndex)
    then result:= TargetURLByIndex(sourceIndex,targetIndex)
    else result:= '';
end;

procedure TApeMappingCatalog.UpdateFromDOM(Doc: TdomDocument);
var
  ape_source,ape_target,ape_title,ape_role: TdomElement;
  targetMappingUrn,sourceHref,targetHref,targetTitle,targetMappingTitle: wideString;
begin
  clear;
  with Doc.documentElement do begin
    if (namespaceURI = xmlns_ape) and (localName='catalog') then begin

      ape_source:= getFirstChildElementNS(xmlns_ape,'source');
      while assigned(ape_source) do begin

        sourceHref:= ape_source.getAttributeNSNormalizedValue(xmlns_ape,'href');
        if sourceHref <> '' then begin

          ape_target:= ape_source.getFirstChildElementNS(xmlns_ape,'target');
          while assigned(ape_target) do begin

            targetHref:= ape_target.getAttributeNSNormalizedValue(xmlns_ape,'href');
            if targetHref <> '' then begin

              ape_title:= ape_target.getFirstChildElementNS(xmlns_ape,'title');
              if assigned(ape_title) then begin
                targetTitle:= ape_title.textContent;

                ape_role:= ape_target.getFirstChildElementNS(xmlns_ape,'role');
                while assigned(ape_role) do begin

                  targetMappingUrn:= ape_role.getAttributeNSNormalizedValue(xmlns_ape,'href');

                  ape_title:= ape_role.getFirstChildElementNS(xmlns_ape,'title');
                  if assigned(ape_title) then begin

                    targetMappingTitle:= ape_title.textContent;
                           // In this version the value of targetMappingTitle
                           // is identical with the value of TargetTitle.
                           // So we need only one of them in the Append()
                           // call.
                    Append(targetMappingUrn,sourceHref,targetHref,targetTitle);

                  end; {if ...}


                  ape_role:= ape_role.getNextSiblingElementNS(xmlns_ape,'role');
                end; {while ...}

              end; {if ...}

            end; {if ...}

            ape_target:= ape_target.getNextSiblingElementNS(xmlns_ape,'target');
          end; {while ...}

        end; {if ...}

        ape_source:= ape_source.getNextSiblingElementNS(xmlns_ape,'source');
      end; {while ...}

    end; {if ...}
  end; {with Doc.documentElement ...}
end;

end.
