unit ApeCatalogDlg;

// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// "http://www.mozilla.org/MPL/"
// 
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
// 
// The Original Code is "ApeCatalogDlg.pas".
// 
// The Initial Developer of the Original Code is Dieter Khler (Heidelberg, 
// Germany, "http://www.philo.de/").
// Portions created by the Initial Developer are Copyright (C) 2002
// Dieter Khler. All Rights Reserved.
// 
// Alternatively, the contents of this file may be used under the terms of
// the GNU General Public License Version 2 or later (the "GPL"), in which
// case the provisions of the GPL are applicable instead of those above.
// If you wish to allow use of your version of this file only under the
// terms of the GPL, and not to allow others to use your version of this
// file under the terms of the MPL, indicate your decision by deleting the
// provisions above and replace them with the notice and other provisions
// required by the GPL. If you do not delete the provisions above, a
// recipient may use your version of this file under the terms of any one
// of the MPL or the GPL.

interface

uses
  ApeCaches, XDOM_2_4, ApeEngines, ApeRepositoryViews,
  Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls, Dialogs,
  Buttons, ExtCtrls, ActnList, StdActns, ComCtrls;

type
  TCatalogDialog = class(TForm)
    ActionList1: TActionList;
    AddFileAction: TFileOpen;
    AddFileBtn: TButton;
    CancelBtn: TButton;
    DeleteBtn: TButton;
    DeleteAction: TAction;
    DisplayMappingTabSheet: TTabSheet;
    DomImpl: TDomImplementation;
    InfoFilesTabSheet: TTabSheet;
    OKBtn: TButton;
    PageControl1: TPageControl;
    Panel1: TPanel;
    ResolveMappingTabSheet: TTabSheet;
    Splitter1: TSplitter;
    StatusMemo: TMemo;
    UndoAllAction: TAction;
    UndoAllBtn: TButton;
    XmlToDomParser: TXmlToDomParser;
    XmlStandardDomReader: TXmlStandardDomReader;
    XmlDocBuilder: TXmlDocBuilder;
    DisplayMappingCatalogView: TApeMappingCatalogView;
    ResolveMappingCatalogView: TApeMappingCatalogView;
    InfoFilesCatalogView: TApeMappingCatalogView;
    procedure AddFileActionAccept(Sender: TObject);
    procedure AddFileActionBeforeExecute(Sender: TObject);
    procedure DeleteActionExecute(Sender: TObject);
    procedure DeleteActionUpdate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure UndoAllActionExecute(Sender: TObject);
    procedure UndoAllActionUpdate(Sender: TObject);
  private
    FDisplayMappingCatalogBackup: TApeMappingCatalog;
    FDisplayMappingModified: boolean;
    FInfoFilesCatalogBackup: TApeMappingCatalog;
    FInfoFilesModified: boolean;
    FMappingRepository: TApeMappingRepository;
    FResolveMappingCatalogBackup: TApeMappingCatalog;
    FResolveMappingModified: boolean;
    function AnalyzeMapping(
      const mappingDoc: TdomDocument): boolean;
    function GetDisplayMappingCatalog: TApeMappingCatalog;
    function GetInfoFilesCatalog: TApeMappingCatalog;
    function GetResolveMappingCatalog: TApeMappingCatalog;
    function LoadMappingFromFile(const Path: TFilename;
      const SysId: wideString): boolean;
    procedure SetMappingRepository(const Value: TApeMappingRepository);
    procedure UpdateDisplayMappingCatalog(const MappingCatalog: TApeMappingCatalog); virtual;
    procedure UpdateInfoFilesCatalog(const MappingCatalog: TApeMappingCatalog); virtual;
    procedure UpdateResolveMappingCatalog(const MappingCatalog: TApeMappingCatalog); virtual;
  public
    function AddMappingUrl(const url: string): boolean; virtual;
    property DisplayMappingCatalog: TApeMappingCatalog read GetDisplayMappingCatalog;
    property DisplayMappingModified: boolean read FDisplayMappingModified;
    property InfoFilesCatalog: TApeMappingCatalog read GetInfoFilesCatalog;
    property InfoFilesModified: boolean read FInfoFilesModified;
    property MappingRepository: TApeMappingRepository read FMappingRepository write SetMappingRepository;
    property ResolveMappingCatalog: TApeMappingCatalog read GetResolveMappingCatalog;
    property ResolveMappingModified: boolean read FResolveMappingModified;
  end;

var
  CatalogDialog: TCatalogDialog;

const
  TIME_OUT:  integer    = 30000; // Timeout for HTTP client (in ms)
  xmlns_ape: wideString = 'http://www.philo.de/ape/test-001';

implementation

{$R *.dfm}

function TCatalogDialog.AnalyzeMapping(const mappingDoc: TdomDocument): boolean;
var
  ape_fileDesc: TdomElement;
  ape_head: TdomElement;
  ape_title: TdomElement;
  ape_titleStmt: TdomElement;
  catalogView: TApeMappingCatalogView;
  PModified: ^boolean;
  sourceMappingURN: wideString;
  targetMappingURN: wideString;
  targetTitle: wideString;
  targetURL: wideString;
begin
  result:= false;
  if assigned(mappingDoc.documentElement) then begin
    with mappingDoc.documentElement do begin

      if (namespaceURI = xmlns_ape) then begin
        if (localName='displayMap') then begin
          catalogView:= DisplayMappingCatalogView;
          PModified:= @FDisplayMappingModified;
        end else if (localName='infoMap') then begin
          catalogView:= InfoFilesCatalogView;
          PModified:= @FInfoFilesModified;
        end else if (localName='resolveMap') then begin
          catalogView:= ResolveMappingCatalogView;
          PModified:= @FResolveMappingModified;
        end else exit;
      end else exit;

      sourceMappingURN:= getAttributeNSNormalizedValue(xmlns_ape,'source');
      if sourceMappingURN = '' then exit;

      targetMappingURN:= getAttributeNSNormalizedValue(xmlns_ape,'target');

      ape_head:= getFirstChildElementNS(xmlns_ape,'head');
      if assigned(ape_head) then begin
        ape_fileDesc:= ape_head.getFirstChildElementNS(xmlns_ape,'fileDesc');
        if assigned(ape_fileDesc) then begin
          ape_titleStmt:= ape_fileDesc.getFirstChildElementNS(xmlns_ape,'titleStmt');
          if assigned(ape_titleStmt) then begin
            ape_title:= ape_titleStmt.getFirstChildElementNS(xmlns_ape,'title');
            targetTitle:= ape_title.textContent;
            targetURL:= mappingDoc.systemId;


            result:= catalogView.Add(targetMappingURN,sourceMappingURN,targetURL,targetTitle);
            if result then PModified^:= true;

          end; {if ...}
        end; {if ...}
      end; {if ...}

    end; {with ...}
  end; {if ...}
end;

function TCatalogDialog.AddMappingUrl(const url: string): boolean;
var
  path: TFilename;
  authority,query,fragment: string;
  uriAnalyzer: TUriStrAnalyzer;
begin
  result:= false;
//  NMHTTP1.Cancel;
  uriAnalyzer:= TUriStrAnalyzer.create;
  try
    uriAnalyzer.setUriReference(url);
    with uriAnalyzer do begin
      if HasUriScheme then begin
        if UriScheme = 'file' then begin
          if UriStrToFilename(url,path,authority,query,fragment)
            then result:= LoadMappingFromFile(path,url)
            ; // xxx else ProcessingErrorLabel.Caption:= SInvalidURLError;
          // xxx ProcessingErrorLabel.visible:= not result;
        end else if UriScheme = 'http' then begin
(*
          with NMHTTP1 do begin
            Tag:= 0;
            TimeOut:= TIME_OUT;
            InputFileMode:= true;
            body:= ExtractFilePath(Application.EXEName) + 'temp\cache2.xml';
            try
//xxx              StopToolButton.Enabled:= true;
              Get(url);
//xxx              StopToolButton.Enabled:= false;
              if Tag = 200  then begin// Tag is set to 200 in the OnSuccess event handler
                if LoadMappingFromFile(body,url) then begin
//xxx                  ProcessingErrorLabel.caption:= '';
//xxx                  ProcessingErrorLabel.visible:= false;
                  result:= true;
                end else begin
//xxx                  ProcessingErrorLabel.caption:= SProcessingError;
//xxx                  ProcessingErrorLabel.visible:= true;
                end;
              end;
            except
//xxx              StopToolButton.Enabled:= false;
              if BeenTimedOut then begin
//xxx                ProcessingErrorLabel.caption:= STimeOutError;
//xxx                ProcessingErrorLabel.visible:= true;
              end else if BeenCanceled then begin
//xxx                ProcessingErrorLabel.caption:= '';
//xxx                ProcessingErrorLabel.visible:= false;
              end else begin
//xxx                ProcessingErrorLabel.caption:= SNetworkError;
//xxx                ProcessingErrorLabel.visible:= true;
              end;
            end; {try ... except ...}
          end; {with NMHTTP1 ...}
*)
        end else begin
//xxx          ProcessingErrorLabel.caption:= SInvalidURISchemeError;
//xxx          ProcessingErrorLabel.visible:= true;
        end;
      end else begin
//xxx        ProcessingErrorLabel.caption:= SInvalidURLError;
//xxx        ProcessingErrorLabel.visible:= true;
      end; {if HasUriScheme ... else ...}
    end; {with ...}
  finally
    uriAnalyzer.free;
  end;
end;

procedure TCatalogDialog.AddFileActionAccept(Sender: TObject);
var
  i: integer;
  Save_Cursor:TCursor;
  uri: wideString;
begin
  Update;

  Save_Cursor := Screen.Cursor;
  Screen.Cursor := crHourGlass; // Show hourglass cursor
  try

    with (Sender as TFileOpen).Dialog.Files do
      for i:= pred(count) downto 0 do begin
        uri:= fileNameToUriStr(Strings[i],[]);
        if AddMappingUrl(uri)
         then StatusMemo.Lines.Add(Format('OK -- "%S" cataloged.',[uri]))
         else StatusMemo.Lines.Add(Format('WARNING -- "%S" could not be cataloged.',[uri]));
      end; {for ...}
  finally
    Screen.Cursor := Save_Cursor; // Restore cursor
  end;
end;

procedure TCatalogDialog.AddFileActionBeforeExecute(Sender: TObject);
begin
  with (Sender as TFileOpen).Dialog do begin
    FileName:= '';
    Options:= [ofHideReadOnly,
               ofAllowMultiSelect,
               ofPathMustExist,
               ofFileMustExist,
               ofEnableSizing,
               ofDontAddToRecent];
  end;
end;

procedure TCatalogDialog.DeleteActionExecute(Sender: TObject);
begin
  if PageControl1.ActivePage = DisplayMappingTabSheet then begin
    if DisplayMappingCatalogView.DeleteFocusedTarget then begin
      FDisplayMappingModified:= true;
      StatusMemo.Lines.Add('Display mapping entry deleted.');
    end;
  end else if PageControl1.ActivePage = ResolveMappingTabSheet then begin
    if ResolveMappingCatalogView.DeleteFocusedTarget then begin
      FResolveMappingModified:= true;
      StatusMemo.Lines.Add('Resolve mapping entry deleted.');
    end;
  end else if PageControl1.ActivePage = InfoFilesTabSheet then begin
    if InfoFilesCatalogView.DeleteFocusedTarget then begin
      FInfoFilesModified:= true;
      StatusMemo.Lines.Add('Info file entry deleted.');
    end;
  end;
end;

procedure TCatalogDialog.DeleteActionUpdate(Sender: TObject);
begin
  if PageControl1.ActivePage = DisplayMappingTabSheet then begin
    (Sender as TAction).Enabled:= DisplayMappingCatalogView.TargetFocused;
  end else if PageControl1.ActivePage = ResolveMappingTabSheet then begin
    (Sender as TAction).Enabled:= ResolveMappingCatalogView.TargetFocused;
  end else if PageControl1.ActivePage = InfoFilesTabSheet then begin
    (Sender as TAction).Enabled:= InfoFilesCatalogView.TargetFocused;
  end;
end;

procedure TCatalogDialog.FormCreate(Sender: TObject);
begin
  FDisplayMappingCatalogBackup:= TApeMappingCatalog.Create;
  FDisplayMappingModified:= false;
  FInfoFilesCatalogBackup:= TApeMappingCatalog.Create;
  FInfoFilesModified:= false;
  FMappingRepository:= nil;
  FResolveMappingCatalogBackup:= TApeMappingCatalog.Create;
  FResolveMappingModified:= false;
end;

procedure TCatalogDialog.FormDestroy(Sender: TObject);
begin
  FDisplayMappingCatalogBackup.free;
  FInfoFilesCatalogBackup.free;
  FResolveMappingCatalogBackup.free;
end;

function TCatalogDialog.GetDisplayMappingCatalog: TApeMappingCatalog;
begin
  result:= DisplayMappingCatalogView.MappingCatalog;
end;

function TCatalogDialog.GetInfoFilesCatalog: TApeMappingCatalog;
begin
  result:= InfoFilesCatalogView.MappingCatalog;
end;

function TCatalogDialog.GetResolveMappingCatalog: TApeMappingCatalog;
begin
  result:= ResolveMappingCatalogView.MappingCatalog;
end;

function TCatalogDialog.LoadMappingFromFile(const Path: TFilename;
                                            const SysId: wideString): boolean;
var
  mappingDocCopy: TdomDocument;
  dummyDoc: TdomDocument;
begin
  result:= false;
  if not fileExists(Path)
    then exit;

  try
    with XmlToDomParser do begin
      dummyDoc:= FileToDom(Path);
      try
        result:= dummyDoc.prepareASModels;
        if result
          then result:= dummyDoc.validate(erReplace);
        if result then begin
          dummyDoc.systemId:= sysId;
          // Parse into namespace aware document tree:
          mappingDocCopy:= DomImpl.createDocument('dummy',nil);
          try
            mappingDocCopy.clear;
            XmlDocBuilder.referenceNode:= mappingDocCopy;
            result:= XmlStandardDomReader.parse(dummyDoc);
            if result then begin
              mappingDocCopy.systemId:= dummyDoc.systemId;
              result:= AnalyzeMapping(mappingDocCopy);
            end;
          finally
            DomImpl.freeDocument(mappingDocCopy);
          end;
        end;
      finally
        DomImpl.freeDocument(dummyDoc);
        DomImpl.freeUnusedASModels;
      end;
    end;
  except
    result:= false;
  end;
end;

procedure TCatalogDialog.SetMappingRepository(const Value: TApeMappingRepository);
begin
  if FMappingRepository <> Value then begin
    FMappingRepository := Value;
    if assigned(FMappingRepository) then begin
      UpdateDisplayMappingCatalog(FMappingRepository.DisplayMappingCatalog);
      UpdateInfoFilesCatalog(FMappingRepository.InfoFilesCatalog);
      UpdateResolveMappingCatalog(FMappingRepository.ResolveMappingCatalog);
    end else begin
      UpdateDisplayMappingCatalog(nil);
      UpdateInfoFilesCatalog(nil);
      UpdateResolveMappingCatalog(nil);
    end;
  end;
end;

procedure TCatalogDialog.UndoAllActionExecute(Sender: TObject);
begin
  if FDisplayMappingModified then UpdateDisplayMappingCatalog(FDisplayMappingCatalogBackup);
  if FInfoFilesModified then UpdateInfoFilesCatalog(FInfoFilesCatalogBackup);
  if FResolveMappingModified then UpdateResolveMappingCatalog(FResolveMappingCatalogBackup);
  with StatusMemo do begin
    Clear;
    Lines.Add('All modifications deprecated.');
  end;
end;

procedure TCatalogDialog.UndoAllActionUpdate(Sender: TObject);
begin
  (Sender as TAction).Enabled:=    FDisplayMappingModified
                                or FInfoFilesModified
                                or FResolveMappingModified;
end;

procedure TCatalogDialog.UpdateDisplayMappingCatalog(const MappingCatalog: TApeMappingCatalog);
begin
  DisplayMappingCatalogView.UpdateView(MappingCatalog);
  FDisplayMappingCatalogBackup.Assign(MappingCatalog);
  FDisplayMappingModified:= false;
end;

procedure TCatalogDialog.UpdateInfoFilesCatalog(
  const MappingCatalog: TApeMappingCatalog);
begin
  InfoFilesCatalogView.UpdateView(MappingCatalog);
  FInfoFilesCatalogBackup.Assign(MappingCatalog);
  FInfoFilesModified:= false;
end;

procedure TCatalogDialog.UpdateResolveMappingCatalog(const MappingCatalog: TApeMappingCatalog);
begin
  ResolveMappingCatalogView.UpdateView(MappingCatalog);
  FResolveMappingCatalogBackup.Assign(MappingCatalog);
  FResolveMappingModified:= false;
end;

end.
