unit ApeUserInterface;

// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// "http://www.mozilla.org/MPL/"
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// The Original Code is "ApeUserInterface.pas".
//
// The Initial Developer of the Original Code is Dieter Khler (Heidelberg,
// Germany, "http://www.philo.de/").
// Portions created by the Initial Developer are Copyright (C) 2002
// Dieter Khler. All Rights Reserved.
//
// Alternatively, the contents of this file may be used under the terms of
// the GNU General Public License Version 2 or later (the "GPL"), in which
// case the provisions of the GPL are applicable instead of those above.
// If you wish to allow use of your version of this file only under the
// terms of the GPL, and not to allow others to use your version of this
// file under the terms of the MPL, indicate your decision by deleting the
// provisions above and replace them with the notice and other provisions
// required by the GPL. If you do not delete the provisions above, a
// recipient may use your version of this file under the terms of any one
// of the MPL or the GPL.

interface

uses
  XDOM_2_4, ApeAboutBox, ApePreferences, ApeUtils, ApeEngines, ApeCatalogDlg,
  CrossReferenceDisplayFrame, ResourceDisplayFrame, StructureDisplayFrame,
  IdBaseComponent, IdAntiFreezeBase, IdAntiFreeze, Windows, Controls, IniFiles,
  ActnList, Menus, Forms, ExtCtrls, SysUtils, Graphics, Classes, ComCtrls;

// The main display components of the User Interface class are the Primary
// Display in the upper left corner, the Secondary Display in the lower left
// corner and the Structure Display at the right side of the form.
//
// The data flow is controled by events of the Structure Display which
// update the Primary and the Secondary Display.  In particular the
// following event handlers are involved:
//
//    StructureDisplayActivePageChange
//    StructureDisplayFocusChanged
//    StructureDisplayInfoModified
//
// If an error occured during a Structur Display operation, the
// StructureDisplayError event handler is used to display error information
// in the status bar.

type
  TUserInterface = class(TForm)
    AboutMenuItem: TMenuItem;
    ActionList: TActionList;
    ApeMappingRepository: TApeMappingRepository;
    ApeMenuItem: TMenuItem;
    ApeUrnRepository: TApeUrnRepository;
    AdnministrationMenuItem: TMenuItem;
    CatalogsMenuItem: TMenuItem;
    EditPreferencesMenuItem: TMenuItem;
    ExitAction: TAction;
    ExitMenueItem: TMenuItem;
    HelpMenuItem: TMenuItem;
    MainMenu: TMainMenu;
    N1: TMenuItem;
    Panel1: TPanel;
    PrimaryDisplay: TApeResourceDisplay;
    ProfilesMenuItem: TMenuItem;
    SecondaryDisplay: TApeCrossReferenceDisplay;
    ShowAboutAction: TAction;
    Splitter1: TSplitter;
    Splitter2: TSplitter;
    StatusBar: TStatusBar;
    StructureDisplay: TApeStructureDisplay;
    IdAntiFreeze1: TIdAntiFreeze;
    ApePreferences: TApePreferences;
    procedure ApePreferencesChange(Sender: TObject);
    procedure ApePreferencesChanging(Sender: TObject;
      UserIndex: Integer; var AllowChange: Boolean);
    procedure CatalogsMenuItemClick(Sender: TObject);
    procedure EditPreferencesMenuItemClick(Sender: TObject);
    procedure ExitActionExecute(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure MappingRepositoryBeginEvaluate(Sender: TObject;
      FileURL: WideString);
    procedure MappingRepositoryEndEvaluate(Sender: TObject;
      FileURL: WideString; Success: Boolean);
    procedure ProfilesMenuItemClick(Sender: TObject);
    procedure ShowAboutActionExecute(Sender: TObject);
    procedure StructureDisplayActivePageChange(Sender: TObject);
    procedure StructureDisplayError(Sender: TObject;
      error: TStructureDisplayError);
    procedure StructureDisplayFocusChanged(Sender: TObject);
    procedure StructureDisplayInfoModified(Sender: TObject);
    procedure FormShow(Sender: TObject);
  protected
    procedure ClearStatusDisplay; virtual;
    procedure HelpMenuOnClick(Sender: TObject); virtual;
  public
    procedure UpdateBoundsRect(const R: TRect);
  end;

var
  UserInterface: TUserInterface;

implementation

const
  DEFAULT_PROFILE_NAME: string = 'default';

{$R *.DFM}

procedure TUserInterface.ApePreferencesChange(Sender: TObject);
begin
  // Updates the catalog at application start and when the profile has changed:
  ApeMappingRepository.UpdateLocalCatalogs;

  // Assign the preferences of the current active user profil to this TUserInterface:
  Assign(Sender as TApePreferences);
end;

procedure TUserInterface.ApePreferencesChanging(Sender: TObject;
  UserIndex: Integer; var AllowChange: Boolean);
begin
  // Update the profile's *.ini file before the active profile changes:
  ApePreferences.Assign(self);
end;

procedure TUserInterface.CatalogsMenuItemClick(Sender: TObject);
begin
  ApeUrnRepository.OpenCatalogAdminDlg;
end;

procedure TUserInterface.EditPreferencesMenuItemClick(Sender: TObject);
begin
  with ApePreferences do
    if ActiveUserIndex > -1
      then if ShowEditProfileDialog(ActiveUserIndex)
        then ApePreferencesChange(ApePreferences);
end;

procedure TUserInterface.ExitActionExecute(Sender: TObject);
begin
  close;  // Closes the application.
end;

procedure TUserInterface.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  // Update the current profile's *.ini file:
  ApePreferences.Assign(self);
end;

procedure TUserInterface.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  CanClose:= StructureDisplay.CloseAllStructures;
end;

procedure TUserInterface.FormCreate(Sender: TObject);
var
  HelpConfig: TStringList;
  I: integer;
  MemIniFile: TMemIniFile;
  P: string;
  Mappings: TStringList;
  Url: WideString;
begin
  try
    // Set template file paths:
    ApePreferences.TemplateDirectory := ExtractFilePath(Application.EXEName) + 'templates\';
    StructureDisplay.TemplateFile:= ApePreferences.TemplateDirectory + 'structure.xml';

    // Set default file of the primary display:
    PrimaryDisplay.DefaultResource:= ApePreferences.TemplateDirectory + '1st.html';

    // Set default file of the secondary display (No default file at the moment):
    SecondaryDisplay.DefaultFile:= ApePreferences.TemplateDirectory + 'empty.html';
    SecondaryDisplay.DefaultResource:= ApePreferences.TemplateDirectory + 'empty.html';

    // Set path to the help file:
    HelpConfig:= TStringList.Create;
    try
      HelpConfig.LoadFromFile(ExtractFilePath(Application.EXEName) + 'help\help.cfg');
      i:= 0;
      while i < HelpConfig.Count do begin
        SecondaryDisplay.AddHelp(ExtractFilePath(Application.EXEName) + 'help\' + HelpConfig[i],HelpConfig[i+1]);
        HelpMenuItem.Insert(i div 2,NewItem(HelpConfig[i+1],0,false,true,HelpMenuOnClick,0,''));
        inc(i,2);
      end;
      SecondaryDisplay.UpdateHelp;
    finally
      HelpConfig.Free;
    end;

    ApePreferences.UsersDirectory:= ExtractFilePath(Application.EXEName) + 'user\';

    MemIniFile:= TMemIniFile.Create(ExtractFilePath(Application.EXEName) + 'settings.ini');
    try
      if MemIniFile.ReadBool('Autorun','on',false) then begin

        // Autorun:

        P := MemIniFile.ReadString('Autorun','profile',DEFAULT_PROFILE_NAME);
        for I := 0 to Pred(ApePreferences.UserCount) do begin
          if ApePreferences.Users[I].ProfileName = P then begin
            ApePreferences.ActiveUserIndex := I;
            Break;
          end;
        end;
        if ApePreferences.ActiveUserIndex = -1
          then raise Exception.CreateFmt('Default profile ''%s'' not found. Autorun failed.',[P]);

        if MemIniFile.ReadBool('Autorun','autoinstall',false) then begin
          Mappings:= TStringList.Create;
          try
            Mappings.Delimiter := ';';
            Mappings.QuoteChar := '"';
            Mappings.DelimitedText := MemIniFile.ReadString('Autorun','mappings','');
            for I := 0 to Pred(Mappings.Count) do begin
              if Trim(Mappings[I]) <> '' then begin
                Url:= fileNameToUriStr(ExtractFilePath(Application.EXEName)+Mappings[I],[]);
                if not ApeMappingRepository.AddMappingUrl(Url)
                  then raise Exception.CreateFmt('Unable to add mapping URL "%s". Autorun failed.', [Url]);
              end;
            end;
          finally
            Mappings.Free;
          end;
          MemIniFile.WriteBool('Autorun','autoinstall',false);
          MemIniFile.UpdateFile; // Save changes.
        end;

      end else begin
        // Select profile:
        case ApePreferences.UserCount of
          1: ApePreferences.ActiveUserIndex:= 0;
        else
          if not ApePreferences.ShowProfileManager(soInitializing)
            then raise Exception.Create('Profile selection failed');
        end;
      end;
    finally
      MemIniFile.free;
    end;
  except
    on E: Exception do begin
      Application.ShowException(E);
      Application.Terminate;
    end;
  end;
end;

procedure TUserInterface.HelpMenuOnClick(Sender: TObject);
begin
  SecondaryDisplay.SwitchToHelp((Sender as TMenuItem).MenuIndex);
end;

procedure TUserInterface.MappingRepositoryBeginEvaluate(Sender: TObject;
  FileURL: WideString);
begin
  StatusBar.SimpleText:= 'Analyzing ' + FileURL;
end;

procedure TUserInterface.MappingRepositoryEndEvaluate(Sender: TObject;
  FileURL: WideString; Success: Boolean);
begin
  StatusBar.SimpleText:= '';
end;

procedure TUserInterface.StructureDisplayActivePageChange(Sender: TObject);
begin
  ClearStatusDisplay;
  StructureDisplayInfoModified(Sender);
end;

procedure TUserInterface.StructureDisplayFocusChanged(Sender: TObject);
var
  UrnList: TdomWideStringList;
  commentText: wideString;
  newDisplayPacket_1,newDisplayPacket_2: TApeDisplayPacket;
begin
  ClearStatusDisplay;
  commentText:= (Sender as TApeStructureDisplay).ActiveCommentText;

  UrnList:= TdomWideStringList.create;
  try
    UrnList.assign((Sender as TApeStructureDisplay).ActiveResourceMap);

    newDisplayPacket_1:= ApeUrnRepository.RequestDisplayPacket(UrnList);
    newDisplayPacket_2:= TApeDisplayPacket.create;
    newDisplayPacket_2.assign(newDisplayPacket_1);

    PrimaryDisplay.Init(newDisplayPacket_1);
    SecondaryDisplay.DisplayComment(commentText);
    SecondaryDisplay.InitApparatus(newDisplayPacket_2);

  finally
    UrnList.free;
  end; {try ...}
end;

procedure TUserInterface.StructureDisplayInfoModified(Sender: TObject);
var
  adp: TApeDisplayPriorityList;
begin
  adp:= (Sender as TApeStructureDisplay).ActiveDisplayPriority;
  if assigned(adp)
    then PrimaryDisplay.DisplayPriority.assign(adp)
    else PrimaryDisplay.DisplayPriority.clear;
  StructureDisplayFocusChanged(Sender);
end;

procedure TUserInterface.ShowAboutActionExecute(Sender: TObject);
var
  AboutBox: TAboutBox;
begin
  AboutBox:= TAboutBox.Create(self);
  try
    AboutBox.DemoMode := ApePreferences.DemoMode;
    AboutBox.ShowModal;
  finally
    AboutBox.free;
  end;
end;

procedure TUserInterface.StructureDisplayError(Sender: TObject;
  error: TStructureDisplayError);
var
  outputStr: string;
begin
  case error of
    sdInvalidURISchemeError: outputStr:= SInvalidURISchemeError;
    sdInvalidURLError:       outputStr:= SInvalidURLError;
    sdLoadingCanceledError:  outputStr:= SLoadingCanceledError;
    sdNetworkError:          outputStr:= SNetworkError;
    sdProcessingError:       outputStr:= SProcessingError;
    sdTimeOutError:          outputStr:= STimeOutError;
  end;
  with StatusBar do begin
    Font.Style:= [fsBold];
    SimpleText:= outputStr;
  end;
end;

procedure TUserInterface.ClearStatusDisplay;
begin
  with StatusBar do begin
    SimpleText:= '';
    Font.Style:= [];
  end;
end;

procedure TUserInterface.ProfilesMenuItemClick(Sender: TObject);
begin
  if ApePreferences.ShowProfileManager(soNormal)
    then ApePreferencesChange(ApePreferences);
end;

procedure TUserInterface.FormShow(Sender: TObject);
begin
  // This is a workaround for a bug in some versions of the IE.  The bug
  // causes an access violation when navigating to an anchor inside an HTML
  // document if the WebBrowser component has only been in the background.
  // To avoid this we bring every WebBrowser component once into the foreground
  // before we start navigating.
  SecondaryDisplay.ShowHelpTab;
  SecondaryDisplay.ShowApparatusTab;
  SecondaryDisplay.ShowCommentTab;
end;

procedure TUserInterface.UpdateBoundsRect(const R: TRect);
begin
  inherited;
end;

end.
