unit ApeVirtualStructures;

// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with 
// the License. You may obtain a copy of the License at 
// "http://www.mozilla.org/MPL/"
// 
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
// 
// The Original Code is "ApeVirtualStructures.pas".
// 
// The Initial Developer of the Original Code is Dieter Khler (Heidelberg, 
// Germany, "http://www.philo.de/").
// Portions created by the Initial Developer are Copyright (C) 2002
// Dieter Khler. All Rights Reserved.
// 
// Alternatively, the contents of this file may be used under the terms of
// the GNU General Public License Version 2 or later (the "GPL"), in which 
// case the provisions of the GPL are applicable instead of those above. 
// If you wish to allow use of your version of this file only under the 
// terms of the GPL, and not to allow others to use your version of this
// file under the terms of the MPL, indicate your decision by deleting the 
// provisions above and replace them with the notice and other provisions 
// required by the GPL. If you do not delete the provisions above, a 
// recipient may use your version of this file under the terms of any one 
// of the MPL or the GPL.

interface

uses
  ApeUtils,
  XDOM_2_4, InfoEditDlg, ItemEditDlg, FileDescDisplayFrame, VirtualTrees,
  Graphics, Controls, ComCtrls, Classes;

type
  TTreeData = record
    Item: TdomElement;
  end;

  PTreeData = ^TTreeData;

  TApeVirtualStructureTree = class;
  TApeStructurePageControl = class;
  TApeStructureReader = class;

  TApeVirtualStructureTree = class(TCustomControl)
  private
    FDisplayFontName: TFontName;
    FOnInfoEdited: TApeNotifyElementEvent;     // called when the info has been edited
    FOnItemEdited: TApeNotifyElementEvent;     // called when an item has been edited
    FOnFocusChanged: TNotifyEvent;             // called when the focus goes to a new node
    procedure UpdateDisplayPriority(const dispPrty: TdomElement);
  protected
    FDisplayPriority: TApeDisplayPriorityList;
    FDomImplementation: TDomImplementation;
    FDomToXmlParser: TDomToXmlParser;
    FDoc: TdomDocument;
    FFileDesc: TdomElement;
    FInfoFrame: TFileDescDispFrame;
    FModified: boolean;
    FStructureReader: TApeStructureReader;
    FVirtualStringTree: TVirtualStringTree;
    FXmlDocBuilder: TXmlDocBuilder;
    procedure DeleteCorrespondingDomTree(Node: PVirtualNode); virtual;
    procedure DoInfoEdited(FileDesc: TdomElement); virtual;
    procedure DoItemEdited(EditedElement: TdomElement); virtual;
    procedure DoFocusChanged; virtual;
    function GetAuthorName: WideString; virtual;
    function GetDisplayPriority: TApeDisplayPriorityList; virtual;
    function GetFocusElement: TdomElement; virtual;
    function GetLanguage: TIso639LanguageCode; virtual;
    function GetShowInfo: boolean; virtual;
    function GetSystemId: wideString; virtual;
    procedure SetAuthorName(const Value: WideString); virtual;
    procedure SetDisplayFontName(const Value: TFontName); virtual;
    procedure SetLanguage(const Value: TIso639LanguageCode); virtual;
    procedure SetShowInfo(const Value: boolean); virtual;
    procedure SetSystemId(const Value: wideString); virtual;
    procedure VirtualStringTreeFocusChanged(Sender: TBaseVirtualTree;
                                            Node: PVirtualNode;
                                            Column: TColumnIndex); virtual;
    procedure VirtualStringTreeGetText(    Sender: TBaseVirtualTree;
                                           Node: PVirtualNode;
                                           Column: TColumnIndex;
                                           TextType: TVSTTextType;
                                       var Text: WideString); virtual;
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
    procedure Add; virtual;
    function CanAdd: boolean; virtual;
    function CanDelete: boolean; virtual;
    function CanDown: boolean; virtual;
    function CanEdit: boolean; virtual;
    function CanIndent: boolean; virtual;
    function CanPullOut: boolean; virtual;
    function CanUp: boolean; virtual;
    procedure Delete; virtual;
    function DocumentAsString: string; virtual;
    function DocumentAsWideString: wideString; virtual;
    procedure Down; virtual;
    procedure Indent; virtual;
    procedure InfoEdit; virtual;
    procedure ItemEdit; virtual;
    procedure PullOut; virtual;
    procedure Up; virtual;
    procedure UpdateStructure(const structureDoc: TdomDocument); virtual;
    property AuthorName: WideString read GetAuthorName write SetAuthorName;
    property DisplayPriority: TApeDisplayPriorityList read GetDisplayPriority;
    property FocusElement: TdomElement read GetFocusElement;
    property Modified: boolean read FModified write FModified;
  published
    property Align;
    property DisplayFontName: TFontName read FDisplayFontName write SetDisplayFontName;
    property Language: TIso639LanguageCode read GetLanguage write SetLanguage default iso639_en;
    property ShowInfo: boolean read GetShowInfo write SetShowInfo default false;
    property SystemId: WideString read GetSystemId write SetSystemId;
    property OnFocusChanged: TNotifyEvent read FOnFocusChanged write FOnFocusChanged;
    property OnInfoEdited: TApeNotifyElementEvent read FOnInfoEdited write FOnInfoEdited;
    property OnItemEdited: TApeNotifyElementEvent read FOnItemEdited write FOnItemEdited;
  end;

  TApeStructurePageControl = class(TCustomControl)
  private
    FDisplayFontName: TFontName;
    FOnChange: TNotifyEvent;               // Occurs after a new tab is selected.
    FOnChanging: TTabChangingEvent;        // Occurs immediately before a new tab is selected.
    FOnInfoModified: TApeNotifyIndexEvent; // Called when the active info has been updated.
    FOnItemModified: TApeNotifyIndexEvent; // Called when the active item has been updated.
    FOnFocusChanged: TApeNotifyIndexEvent; // Called after the focus of an item changed.
    procedure PageControlChange(Sender: TObject);
    procedure PageControlChanging(Sender: TObject;
                                  var AllowChange: Boolean);
  protected
    FLanguage: TIso639LanguageCode;
    FPageControl: TPageControl;
    FShowInfo: boolean;
    procedure DoChange; virtual;
    procedure DoChanging(var AllowChange: boolean); virtual;
    procedure DoFocusChanged(PageIndex: integer); virtual;
    procedure DoInfoModified(PageIndex: integer;
                             FileDesc: TdomElement); virtual;
    procedure DoItemModified(PageIndex: integer;
                             ItemElement: TdomElement); virtual;
    function GetActivePage: TTabSheet; virtual;
    function GetActivePageIndex: integer; virtual;
    function GetAuthorName(index: integer): WideString; virtual;
    function GetDisplayPriority(index: integer): TApeDisplayPriorityList; virtual;
    function GetFocusElement(index: integer): TdomElement; virtual;
    function GetIndexOfVirtualStructureTree(vst: TApeVirtualStructureTree): integer; virtual;
    function GetModified(index: integer): boolean; virtual;
    function GetModifiedCount: integer; virtual;
    function GetPageCount: integer; virtual;
    function GetSystemId(index: integer): WideString; virtual;
    function GetVirtualStructureTree(const index: integer): TApeVirtualStructureTree; virtual;
    procedure SetActivePageIndex(const Value: integer); virtual;
    procedure SetAuthorName(index: integer; const Value: WideString); virtual;
    procedure SetDisplayFontName(const Value: TFontName); virtual;
    procedure SetLanguage(const Value: TIso639LanguageCode); virtual;
    procedure SetModified(index: integer; const Value: boolean); virtual;
    procedure SetShowInfo(const Value: boolean); virtual;
    procedure SetSystemId(index: integer; const Value: WideString); virtual;
    procedure UpdateActiveTagCaption(const FileDesc: TdomElement); virtual;
    procedure VSTFocusChanged(Sender: TObject); virtual;
    procedure VSTInfoEdited(Sender: TObject; FileDesc: TdomElement); virtual;
    procedure VSTItemEdited(Sender: TObject; EditedElement: TdomElement); virtual;
  public
    constructor Create(aOwner: TComponent); override;
    function Add: boolean; virtual;
    function AddStructure(const structureDoc: TdomDocument): boolean; virtual;
    function CanAdd: boolean; virtual;
    function CanDelete: boolean; virtual;
    function CanDown: boolean; virtual;
    function CanEdit: boolean; virtual;
    function CanIndent: boolean; virtual;
    function CanPullOut: boolean; virtual;
    function CanUp: boolean; virtual;
    function CloseStructure(const index: integer): boolean; virtual;
    function Delete: boolean; virtual;
    function DocumentAsString(const index: integer): string; virtual;
    function DocumentAsWideString(const index: integer): wideString; virtual;
    function Down: boolean; virtual;
    function Indent: boolean; virtual;
    function InfoEdit: boolean; virtual;
    function ItemEdit: boolean; virtual;
    function PullOut: boolean; virtual;
    function Up: boolean; virtual;
    property ActivePageIndex: integer read GetActivePageIndex write SetActivePageIndex;
    property AuthorName[index: integer]: WideString read GetAuthorName write SetAuthorName;
    property DisplayPriority[index: integer]: TApeDisplayPriorityList read GetDisplayPriority;
    property FocusElement[index: integer]: TdomElement read GetFocusElement;
    property Modified[index: integer]: boolean read GetModified write SetModified;
    property ModifiedCount: integer read GetModifiedCount;
    property PageCount: integer read GetPageCount;
    property SystemId[index: integer]: WideString read GetSystemId write SetSystemId;
  published
    property Align;
    property DisplayFontName: TFontName read FDisplayFontName write SetDisplayFontName;
    property Language: TIso639LanguageCode read FLanguage write SetLanguage default iso639_en;
    property ShowInfo: boolean read FShowInfo write SetShowInfo default false;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnChanging: TTabChangingEvent read FOnChanging write FOnChanging;
    property OnFocusChanged: TApeNotifyIndexEvent read FOnFocusChanged write FOnFocusChanged;
    property OnInfoModified: TApeNotifyIndexEvent read FOnInfoModified write FOnInfoModified;
    property OnItemModfied: TApeNotifyIndexEvent read FOnItemModified write FOnItemModified;
  end;

  TApeStructureReader = class (TXmlStandardDomReader)
  protected
    function parseloopItem(const BVT: TBaseVirtualTree;
                           const Node: PVirtualNode): boolean; virtual;
  public
    function parse(const vst: TApeVirtualStructureTree): boolean; reintroduce; 
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Ape', [TApeVirtualStructureTree,
                             TApeStructurePageControl
                            ]);
end;

{ TApeVirtualStructureTree }

constructor TApeVirtualStructureTree.Create(aOwner: TComponent);
begin
  inherited;

  FModified:= false;

  FDomImplementation:= TDomImplementation.Create(self);
  FDoc:= FDomImplementation.createDocumentNS(xmlns_ape,'dummy',nil);
  FDoc.clear;
  FFileDesc:= nil;

  FDomToXmlParser:= TDomToXmlParser.create(self);
  FDomToXmlParser.DomImpl:= FDomImplementation;

  FXmlDocBuilder:= TXmlDocBuilder.Create(self);
  FXmlDocBuilder.BuildNamespaceTree:= true;
  FStructureReader:= TApeStructureReader.Create(self);
  FStructureReader.NextHandler:= FXmlDocBuilder;
  FStructureReader.DOMImpl:= FDomImplementation;

  FVirtualStringTree:= TVirtualStringTree.Create(self);
  FVirtualStringTree.Parent:= self;
  FVirtualStringTree.Align:= alClient;
  FVirtualStringTree.OnGetText:= VirtualStringTreeGetText;
  FVirtualStringTree.OnFocusChanged:= VirtualStringTreeFocusChanged;
  // Suppress automatic horizontal scrolling:
  FVirtualStringTree.TreeOptions.AutoOptions:=
    FVirtualStringTree.TreeOptions.AutoOptions + [toDisableAutoscrollOnFocus];

  FInfoFrame:= TFileDescDispFrame.Create(self);
  FInfoFrame.Parent:= self;
  FInfoFrame.Align:= alClient;

  FDisplayPriority:= TApeDisplayPriorityList.Create;

  FVirtualStringTree.ParentFont:= false;
  DisplayFontName:= 'MS Sans Serif';  // Automatically sets the display font of all relevant controls.

  Language:= iso639_en;
  ShowInfo:= false;
end;

destructor TApeVirtualStructureTree.Destroy;
begin
  FDisplayPriority.Free;
  inherited;
end;

procedure TApeVirtualStructureTree.DeleteCorrespondingDomTree(
  Node: PVirtualNode);
var
  oldElement: TdomNode;
  data: PTreeData;
begin
  with FVirtualStringTree do begin
    data:= GetNodeData(Node);
    oldElement:= data.Item;
    if assigned(oldElement) then begin
      with oldElement do begin
        if assigned(parentNode)
          then parentNode.removeChild(oldElement);
      end; {with ...}
      FDoc.freeAllNodes(oldElement);
    end; {if ...}
  end; {with ...}
end;

procedure TApeVirtualStructureTree.DoInfoEdited(FileDesc: TdomElement);

  function GetDisplayPriorityNode(FileDesc: TdomElement): TdomElement;
  begin
    result:= FileDesc;
    if assigned(result) then begin
      result:= result.getFirstChildElementNS(xmlns_ape,'renderingStmt');
      if assigned(result) then begin
        result:= result.getFirstChildElementNS(xmlns_ape,'displayPriority');
      end;
    end;
  end;

begin
  UpdateDisplayPriority(GetDisplayPriorityNode(FileDesc));
  if Assigned(FOnInfoEdited) then
    FOnInfoEdited(Self,FileDesc);
end;

procedure TApeVirtualStructureTree.DoItemEdited(EditedElement: TdomElement);
begin
  if Assigned(FOnItemEdited) then
    FOnItemEdited(Self,EditedElement);
end;

procedure TApeVirtualStructureTree.DoFocusChanged;
begin
  if Assigned(FOnFocusChanged) then
    FOnFocusChanged(Self);
end;

procedure TApeVirtualStructureTree.UpdateDisplayPriority(
  const dispPrty: TdomElement);
var
  dispRole: TdomElement;
  uid: wideString;
begin
  FDisplayPriority.clear;
  if assigned(dispPrty) then begin
    dispRole:= dispPrty.getFirstChildElementNS(xmlns_ape,'displayRole');
    while assigned(dispRole) do begin
      uid:= dispRole.getAttributeNSNormalizedValue(xmlns_ape,'uid');
      if uid <> ''
        then FDisplayPriority.Add(uid);
      dispRole:= dispRole.getNextSiblingElementNS(xmlns_ape,'displayRole');
    end;
  end;
end;

procedure TApeVirtualStructureTree.VirtualStringTreeFocusChanged(
  Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex);
begin
  DoFocusChanged;
end;

procedure TApeVirtualStructureTree.VirtualStringTreeGetText(
  Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex;
  TextType: TVSTTextType; var Text: WideString);
var
  Data: PTreeData;
  ape_item, ape_caption: TdomElement;
begin
  data:= Sender.GetNodeData(Node);
  ape_item:= data.item;
  if assigned(ape_item) then begin
    ape_caption:= ape_item.getFirstChildElementNS(xmlns_ape,'caption');
    if assigned(ape_caption)
      then text:= ape_caption.textContent
      else text:= '';
  end else text:= '';
end;

procedure TApeVirtualStructureTree.Add;
var
  tn,lastFocused: PVirtualNode;
  newData: PTreeData;
  newItemElement: TdomElement;
begin
  with FVirtualStringTree do begin
    lastFocused:= FocusedNode;
    if assigned(lastFocused) then begin
      tn:= InsertNode(lastFocused,amAddChildLast);

      newItemElement:= FDoc.createElementNS(xmlns_ape,'ape:item');
      newItemElement.appendChild(FDoc.createElementNS(xmlns_ape,'ape:caption'));
      newItemElement.appendChild(FDoc.createElementNS(xmlns_ape,'ape:comment'));
      newItemElement.appendChild(FDoc.createElementNS(xmlns_ape,'ape:resourceMap'));

      newData:= GetNodeData(tn);
      newData.Item:= newItemElement;
      FocusedNode:= tn;
      Selected[lastFocused]:= False;
      Selected[tn]:= True;

      FModified:= True;
    end;
    ItemEdit;
  end; {with ...}
end;

function TApeVirtualStructureTree.CanAdd: boolean;
begin
  if assigned(FVirtualStringTree.FocusedNode)
    then result:= true
    else result:= false;
end;

function TApeVirtualStructureTree.CanDelete: boolean;
var
  tn: PVirtualNode;
begin
  tn:= FVirtualStringTree.FocusedNode;
  if assigned(tn) then begin
    if FVirtualStringTree.GetNodeLevel(tn) > 0
      then result:= true
      else result:= false;
  end else result:= false;
end;

function TApeVirtualStructureTree.CanDown: boolean;
var
  tn: PVirtualNode;
begin
  tn:= FVirtualStringTree.FocusedNode;
  if assigned(tn) then begin
    if assigned(tn.nextSibling)
      then result:= true
      else result:= false;
  end else result:= false;
end;

function TApeVirtualStructureTree.CanEdit: boolean;
begin
  if assigned(FVirtualStringTree.FocusedNode)
    then result:= true
    else result:= false;
end;

function TApeVirtualStructureTree.CanIndent: boolean;
var
  tn: PVirtualNode;
begin
  tn:= FVirtualStringTree.FocusedNode;
  if assigned(tn) then begin
    if assigned(tn.NextSibling)
      then result:= true
      else result:= false;
  end else result:= false;
end;

function TApeVirtualStructureTree.CanPullOut: boolean;
var
  tn: PVirtualNode;
begin
  tn:= FVirtualStringTree.FocusedNode;
  if assigned(tn) then begin
    if FVirtualStringTree.GetNodeLevel(tn) > 1
      then result:= true
      else result:= false;
  end else result:= false;
end;

function TApeVirtualStructureTree.CanUp: boolean;
var
  tn: PVirtualNode;
begin
  tn:= FVirtualStringTree.FocusedNode;
  if assigned(tn) then begin
    if assigned(tn.prevSibling)
      then result:= true
      else result:= false;
  end else result:= false;
end;

procedure TApeVirtualStructureTree.Delete;
var
  fn,tn: PVirtualNode;
begin
  with FVirtualStringTree do begin
    BeginUpdate;
    try
      tn:= FocusedNode;
      fn:= tn.PrevSibling;
      if not assigned(fn)
        then fn:= tn.NextSibling;
      if assigned(tn) then begin
        FocusedNode:= nil;
        DeleteCorrespondingDomTree(tn);
        DeleteNode(tn);
      end; {if ...}
      FocusedNode:= fn;
      Selected[fn]:= True;
      FModified:= True;
    finally
      EndUpdate;
    end;
  end; {with ...}
end;

function TApeVirtualStructureTree.DocumentAsString: string;
var
  doc: TdomDocument;
begin
  result:= '';
  doc:= FDomImplementation.createDocument('ape:structureTable',nil);
  try
    doc.clear;
    FXmlDocBuilder.referenceNode:= doc;
    if FStructureReader.parse(self)
      then FDomToXmlParser.writeToString(doc,'UTF-8',result)
      else result:= '';
  finally
    FDomImplementation.freeDocument(doc);
  end;
end;

function TApeVirtualStructureTree.DocumentAsWideString: wideString;
var
  doc: TdomDocument;
begin
  doc:= FDomImplementation.createDocument('ape:structureTable',nil);
  try
    doc.clear;
    FXmlDocBuilder.referenceNode:= doc;
    if FStructureReader.parse(self)
      then FDomToXmlParser.writeToWideString(doc,result)
      else result:= '';
  finally
    FDomImplementation.freeDocument(doc);
  end;
end;

procedure TApeVirtualStructureTree.Down;
var
  tn,next_tn: PVirtualNode;
begin
  with FVirtualStringTree do begin
    BeginUpdate;
    try
      tn:= FocusedNode;
      if assigned(tn) then begin
        next_tn:= tn.nextSibling;
        if assigned(next_tn)
          then MoveTo(tn,next_tn,amInsertAfter,false);
      end; {if ...}
      FocusedNode:= tn;
      ScrollIntoView(tn,true,false);
      FModified:= True;
    finally
      EndUpdate;
    end;
  end; {with ...}
end;

procedure TApeVirtualStructureTree.Indent;
var
  tn,nextSibling_tn: PVirtualNode;
begin
  with FVirtualStringTree do begin
    BeginUpdate;
    try
      tn:= FocusedNode;
      if assigned(tn) then begin
        nextSibling_tn:= tn.NextSibling;
        if assigned(nextSibling_tn)
          then MoveTo(tn,nextSibling_tn,amAddChildFirst,false);
      end; {if ...}
      FocusedNode:= tn;
      ScrollIntoView(tn,true,false);
      FModified:= True;
    finally
      EndUpdate;
    end;
  end; {with ...}
end;

procedure TApeVirtualStructureTree.InfoEdit;
var
  InfoEditDlg: TInfoEditDialog;
  oldFileDesc: TdomElement;
begin
  with FVirtualStringTree do begin
    oldFileDesc:= FFileDesc;
    InfoEditDlg:= TInfoEditDialog.Create(self);
    try
      with InfoEditDlg do begin
        InitializeFileDesc(oldFileDesc);
        DisplayFontName:= self.DisplayFontName;
        if ShowModal = mrOK then begin
          FFileDesc:= (FDoc.importNode(FileDescElement,true) as TdomElement); // Copy FileDesc node tree from InfoEditDlg
          FDoc.freeAllNodes(TdomNode(oldFileDesc));                           // Free the old FileDesc
          FModified:= True;
          FInfoFrame.UpdateDisplay(FFileDesc);
          DoInfoEdited(FFileDesc);
        end;
      end;
    finally
      InfoEditDlg.free;
    end;
  end;
end;

procedure TApeVirtualStructureTree.ItemEdit;
var
  ItemEditDlg: TItemEditDialog;
  newItem,oldItem: TdomElement;
  tn: PVirtualNode;
  data: PTreeData;
begin
  with FVirtualStringTree do begin
    tn:= FocusedNode;
    if not assigned(tn) then exit;
    data:= GetNodeData(tn);
    oldItem:= data.Item;
    ItemEditDlg:= TItemEditDialog.Create(self);
    try
      with ItemEditDlg do begin
        InitializeItem(oldItem);
        DisplayFontName:= self.DisplayFontName;
        if ShowModal = mrOK then begin
          newItem:= (FDoc.importNode(ItemElement,true) as TdomElement); // Copy Item node tree from ItemEditDlg
          data.Item:= newItem;                                          // Update the tree view
          FDoc.freeAllNodes(TdomNode(oldItem));                         // Free the old FileDesc
          InvalidateNode(tn);
          FModified:= True;
          DoItemEdited(newItem);
        end; {if ShowModal = mrOK ...}
      end; {with ItemEditDlg ...}
    finally
      ItemEditDlg.free;
    end; {try ... finally ...}
  end; {with FVirtualStringTree ...}
end;

function TApeVirtualStructureTree.GetAuthorName: WideString;
var
  ape_titleStmt,ape_author: TdomElement;
begin
  result:= '';
  if assigned(FFileDesc) then begin
    ape_titleStmt:= FFileDesc.getFirstChildElementNS(xmlns_ape,'titleStmt');
    if assigned(ape_titleStmt) then begin
      ape_author:= ape_titleStmt.getFirstChildElementNS(xmlns_ape,'author');
      if assigned(ape_author) then begin
        result:= ape_author.textcontent;
      end;
    end;
  end;
end;

function TApeVirtualStructureTree.GetDisplayPriority: TApeDisplayPriorityList;
begin
  result:= FDisplayPriority;
end;

function TApeVirtualStructureTree.GetFocusElement: TdomElement;
var
  Data: PTreeData;
begin
  with FVirtualStringTree do begin
    if assigned(FocusedNode) then begin
      data:= GetNodeData(FocusedNode);
      result:= data.item;
    end else result:= nil;
  end;
end;

function TApeVirtualStructureTree.GetLanguage: TIso639LanguageCode;
begin
  result:= FInfoFrame.Language;
end;

function TApeVirtualStructureTree.GetShowInfo: boolean;
begin
  result:= FInfoFrame.Visible;
end;

function TApeVirtualStructureTree.GetSystemId: wideString;
begin
  result:= FDoc.systemId;
end;

procedure TApeVirtualStructureTree.PullOut;
var
  tn,parent_tn: PVirtualNode;
begin
  with FVirtualStringTree do begin
    BeginUpdate;
    try
      tn:= FocusedNode;
      if GetNodeLevel(tn) > 1 then begin
        parent_tn:= tn.Parent;
        if assigned(parent_tn)
          then MoveTo(tn,parent_tn,amInsertBefore,false);
      end; {if ...}
      FocusedNode:= tn;
      ScrollIntoView(tn,true,false);
      FModified:= True;
    finally
      EndUpdate;
    end;
  end; {with ...}
end;

procedure TApeVirtualStructureTree.SetAuthorName(const Value: WideString);
var
  ape_titleStmt,ape_author: TdomElement;
begin
  if assigned(FFileDesc) then begin
    ape_titleStmt:= FFileDesc.getFirstChildElementNS(xmlns_ape,'titleStmt');
    if assigned(ape_titleStmt) then begin
      ape_author:= ape_titleStmt.getFirstChildElementNS(xmlns_ape,'author');
      if assigned(ape_author) then begin
        with ape_author do begin
          clear;
          appendChild(ownerDocument.createTextNode(Value));
        end;
      end;
    end;
  end;
  FInfoFrame.UpdateDisplay(FFileDesc);
end;

procedure TApeVirtualStructureTree.SetDisplayFontName(const Value: TFontName);
begin
  if FDisplayFontName <> Value then begin
    FDisplayFontName:= Value;
    FVirtualStringTree.Font.Name:= Value;
    FInfoFrame.DisplayFontName:= Value;
  end;
end;

procedure TApeVirtualStructureTree.SetLanguage(
  const Value: TIso639LanguageCode);
begin
  FInfoFrame.Language:= Value;
end;

procedure TApeVirtualStructureTree.SetShowInfo(const Value: boolean);
begin
  FInfoFrame.Visible:= Value;
  FVirtualStringTree.Visible:= not Value;
end;

procedure TApeVirtualStructureTree.SetSystemId(const Value: wideString);
begin
  FDoc.systemId:= Value;
end;

procedure TApeVirtualStructureTree.Up;
var
  tn,previous_tn: PVirtualNode;
begin
  with FVirtualStringTree do begin
    BeginUpdate;
    try
      tn:= FocusedNode;
      if assigned(tn) then begin
        previous_tn:= tn.PrevSibling;
        if assigned(previous_tn)
          then MoveTo(tn,previous_tn,amInsertBefore,false);
      end; {if ...}
      FocusedNode:= tn;
      ScrollIntoView(tn,true,false);
      FModified:= True;
    finally
      EndUpdate;
    end;
  end; {with ...}
end;

procedure TApeVirtualStructureTree.UpdateStructure(
  const structureDoc: TdomDocument);
var
  otherFileDesc: TdomElement;

  function GetFileDescNode(const doc: TdomDocument): TdomElement;
  var
    ape_APE: TdomElement;
    ape_head: TdomElement;
  begin
    result:= nil;
    if assigned(doc) then begin
      ape_APE:= doc.getFirstChildElementNS(xmlns_ape,'structureTable');
      if assigned(ape_APE) then begin
        ape_head:= ape_APE.getFirstChildElementNS(xmlns_ape,'head');
        if assigned(ape_head)
          then result:= ape_head.getFirstChildElementNS(xmlns_ape,'fileDesc');
      end;
    end;
  end;

  function GetDisplayPriorityNode(const doc: TdomDocument): TdomElement;
  begin
    result:= GetFileDescNode(doc);
    if assigned(result) then begin
      result:= result.getFirstChildElementNS(xmlns_ape,'renderingStmt');
      if assigned(result) then begin
        result:= result.getFirstChildElementNS(xmlns_ape,'displayPriority');
      end;
    end;
  end;

  procedure HandleNodeList(pn: PVirtualNode; DomNodeList:TDomNodeList);
  var
    i:integer;
    ape_item_Int, ape_caption_Ext, ape_caption_Int: TdomElement;
    ape_comment_Ext, ape_comment_Int: TdomElement;
    ape_resourceMap_Ext, ape_resourceMap_Int: TdomElement;
    DomNode:TDomNode;
    tn: PVirtualNode;
    Data: PTreeData;
  begin
    tn:= pn;
    for i:=0 to pred(DomNodeList.length) do begin
      DomNode:= DomNodeList.Item(i);
      with DomNode do begin
        case nodeType of
          ntElement_Node: begin
            if namespaceURI = xmlns_ape then begin
              if localName = 'item' then begin
                tn:= FVirtualStringTree.AddChild(pn);
                data:= FVirtualStringTree.GetNodeData(tn);
                ape_item_Int:= FDoc.createElementNS(xmlns_ape,'ape:item');
                data.item:= ape_item_Int;

                // caption
                ape_caption_Ext:= domNode.getFirstChildElementNS(xmlns_ape,'caption');
                if assigned(ape_caption_Ext)
                  then ape_caption_Int:= (FDoc.importNode(ape_caption_Ext,true) as TdomElement)
                  else ape_caption_Int:= FDoc.createElementNS(xmlns_ape,'ape:caption');
                ape_item_Int.appendChild(ape_caption_Int);

                // comment
                ape_comment_Ext:= domNode.getFirstChildElementNS(xmlns_ape,'comment');
                if assigned(ape_comment_Ext)
                  then ape_comment_Int:= (FDoc.importNode(ape_comment_Ext,true) as TdomElement)
                  else ape_comment_Int:= FDoc.createElementNS(xmlns_ape,'ape:comment');
                ape_item_Int.appendChild(ape_comment_Int);

                // resourceMap
                ape_resourceMap_Ext:= domNode.getFirstChildElementNS(xmlns_ape,'resourceMap');
                if assigned(ape_resourceMap_Ext)
                  then ape_resourceMap_Int:= (FDoc.importNode(ape_resourceMap_Ext,true) as TdomElement)
                  else ape_resourceMap_Int:= FDoc.createElementNS(xmlns_ape,'ape:resourceMap');
                ape_item_Int.appendChild(ape_resourceMap_Int);

              end;
            end;
          end;
        end;
      end; {with ...}
      if assigned(DomNode.ChildNodes)
        then HandleNodeList(tn, DomNode.ChildNodes);
    end;
  end;

begin
  with FVirtualStringTree do begin
    Clear;
    FDoc.clear;
    NodeDataSize := SizeOf(TTreeData);
    if assigned(structureDoc)
      then HandleNodeList(nil,structureDoc.childNodes);
    FModified:= False;
    FullExpand;
  end; {with ...}

  UpdateDisplayPriority(GetDisplayPriorityNode(structureDoc));

  otherFileDesc:= GetFileDescNode(structureDoc);
  if assigned(otherFileDesc)
    then FFileDesc:= (FDoc.importNode(otherFileDesc,true) as TdomElement) // Copy FileDesc node tree from structureDoc
    else FFileDesc:= nil;
  FInfoFrame.UpdateDisplay(FFileDesc);

  DoFocusChanged;
end;



{ TApeStructurePageControl }

constructor TApeStructurePageControl.Create(aOwner: TComponent);
begin
  inherited;

  FPageControl:= TPageControl.Create(self);
  with FPageControl do begin
    Parent:= self;
    Align:= alClient;
    OnChange:= PageControlChange;
    OnChanging:= PageControlChanging;
  end;

  FDisplayFontName:= 'MS Sans Serif';
  FLanguage:= iso639_en;
  FShowInfo:= false;
end;

procedure TApeStructurePageControl.DoChange;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TApeStructurePageControl.DoChanging(var AllowChange: boolean);
begin
  if Assigned(FOnChanging) then
    FOnChanging(Self,AllowChange);
end;

procedure TApeStructurePageControl.DoFocusChanged(PageIndex: integer);
begin
  if Assigned(FOnFocusChanged) then
    FOnFocusChanged(Self,PageIndex);
end;

procedure TApeStructurePageControl.DoInfoModified(PageIndex: integer;
                                                  FileDesc: TdomElement);
begin
  UpdateActiveTagCaption(FileDesc);
  if Assigned(FOnInfoModified) then
    FOnInfoModified(Self,PageIndex);
end;

procedure TApeStructurePageControl.DoItemModified(PageIndex: integer;
                                                  ItemElement: TdomElement);
begin
  if Assigned(FOnItemModified) then
    FOnItemModified(Self,PageIndex);
end;

function TApeStructurePageControl.GetActivePage: TTabSheet;
begin
  result:= FPageControl.ActivePage;
end;

function TApeStructurePageControl.GetActivePageIndex: integer;
begin
  result:= FPageControl.ActivePageIndex;
end;

function TApeStructurePageControl.GetAuthorName(index: integer): WideString;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(index);
  if assigned(avst)
    then result:= avst.AuthorName
    else result:= '';
end;

function TApeStructurePageControl.GetDisplayPriority(
  index: integer): TApeDisplayPriorityList;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(index);
  if assigned(avst)
    then result:= avst.DisplayPriority
    else result:= nil;
end;

function TApeStructurePageControl.GetFocusElement(
  index: integer): TdomElement;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(index);
  if assigned(avst)
    then result:= avst.FocusElement
    else result:= nil;
end;

function TApeStructurePageControl.GetIndexOfVirtualStructureTree(
  vst: TApeVirtualStructureTree): integer;
var
  i: integer;
begin
  result:= -1;
  for i:= 0 to FPageControl.PageCount do
    if GetVirtualStructureTree(i) = vst then begin
      result:= i;
      exit;
    end;
end;

function TApeStructurePageControl.GetModified(index: integer): boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(index);
  if assigned(avst)
    then result:= avst.Modified
    else result:= false;
end;

function TApeStructurePageControl.GetModifiedCount: integer;
var
  i: integer;
begin
  result:= 0;
  for i:= 0 to pred(PageCount) do
    if Modified[i] then inc(result);
end;

function TApeStructurePageControl.GetPageCount: integer;
begin
  result:= FPageControl.PageCount;
end;

function TApeStructurePageControl.GetSystemId(index: integer): WideString;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(index);
  if assigned(avst)
    then result:= avst.SystemId
    else result:= '';
end;

function TApeStructurePageControl.GetVirtualStructureTree(const index: integer): TApeVirtualStructureTree;
var
  i: integer;
begin
  result:= nil;
  // Out of bounds?
  if (index < 0) or (index > pred(FPageControl.PageCount)) then exit;
  with FPageControl.Pages[index] do begin
    for i:= 0 to pred(ControlCount) do begin
      if Controls[i] is TApeVirtualStructureTree then begin
        result:= (Controls[i] as TApeVirtualStructureTree);
        exit;
      end;
    end;
  end;
end;

procedure TApeStructurePageControl.PageControlChange(Sender: TObject);
begin
  DoChange;
end;

procedure TApeStructurePageControl.PageControlChanging(Sender: TObject;
  var AllowChange: Boolean);
begin
  DoChanging(AllowChange);
end;

procedure TApeStructurePageControl.SetActivePageIndex(
  const Value: integer);
begin
  FPageControl.ActivePageIndex:= Value;
end;

procedure TApeStructurePageControl.SetAuthorName(index: integer;
  const Value: WideString);
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(index);
  if assigned(avst)
    then avst.AuthorName:= Value;
end;

procedure TApeStructurePageControl.SetDisplayFontName(
  const Value: TFontName);
var
  i: integer;
begin
  if FDisplayFontName <> Value then begin
    for i:= 0 to pred(PageCount) do
      GetVirtualStructureTree(i).DisplayFontName:= Value;
    FDisplayFontName:= Value;
  end;
end;

procedure TApeStructurePageControl.SetLanguage(
  const Value: TIso639LanguageCode);
var
  i: integer;
begin
  if FLanguage <> Value then begin
    for i:= 0 to pred(PageCount) do
      GetVirtualStructureTree(i).Language:= Value;
    FLanguage:= Value;
  end;
end;

procedure TApeStructurePageControl.SetModified(index: integer;
  const Value: boolean);
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(index);
  if assigned(avst)
    then avst.Modified:= Value;
end;

procedure TApeStructurePageControl.SetShowInfo(const Value: boolean);
var
  i: integer;
begin
  if FShowInfo <> Value then begin
    for i:= 0 to pred(PageCount) do
      GetVirtualStructureTree(i).ShowInfo:= Value;
    FShowInfo:= Value;
  end;
end;

procedure TApeStructurePageControl.SetSystemId(index: integer;
  const Value: WideString);
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(index);
  if assigned(avst)
    then avst.SystemId:= Value;
end;

procedure TApeStructurePageControl.UpdateActiveTagCaption(
  const FileDesc: TdomElement);
var
  ape_titleStmt,ape_title: TdomElement;
  tabText: wideString;
begin
  if assigned(FileDesc) then begin
    tabText:= '';
    ape_titleStmt:= FileDesc.getFirstChildElementNS(xmlns_ape,'titleStmt');
    if assigned(ape_titleStmt) then begin
      ape_title:= ape_titleStmt.getFirstChildElementNS(xmlns_ape,'title');
      while assigned(ape_title) do begin
        if ape_title.getAttributeNSNormalizedValue(xmlns_ape,'type') = 'label' then begin
          tabText:= ape_title.textContent;
          break;
        end;
        ape_title:= ape_title.getNextSiblingElementNS(xmlns_ape,'title');
      end; {if ...}
    end; {if ...}
    FPageControl.ActivePage.Caption:= tabText;
  end;
end;

procedure TApeStructurePageControl.VSTFocusChanged(
  Sender: TObject);
begin
  DoFocusChanged(GetIndexOfVirtualStructureTree(Sender as TApeVirtualStructureTree));
end;

procedure TApeStructurePageControl.VSTInfoEdited(
  Sender: TObject; FileDesc: TdomElement);
begin
  DoInfoModified(GetIndexOfVirtualStructureTree(Sender as TApeVirtualStructureTree),
                 FileDesc);
end;

procedure TApeStructurePageControl.VSTItemEdited(
  Sender: TObject; EditedElement: TdomElement);
begin
  DoItemModified(GetIndexOfVirtualStructureTree(Sender as TApeVirtualStructureTree),
                 EditedElement);
end;

function TApeStructurePageControl.Add: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then begin avst.Add; result:= true; end
    else result:= false;
end;

function TApeStructurePageControl.AddStructure(
  const structureDoc: TdomDocument): boolean;
var
  newTabSheet: TTabSheet;

  function GetFileDescNode(const doc: TdomDocument): TdomElement;
  var
    ape_APE: TdomElement;
    ape_head: TdomElement;
  begin
    result:= nil;
    if assigned(doc) then begin
      ape_APE:= doc.getFirstChildElementNS(xmlns_ape,'structureTable');
      if assigned(ape_APE) then begin
        ape_head:= ape_APE.getFirstChildElementNS(xmlns_ape,'head');
        if assigned(ape_head)
          then result:= ape_head.getFirstChildElementNS(xmlns_ape,'fileDesc');
      end;
    end;
  end;

begin
  if assigned(structureDoc)
    then result:= true
    else begin result:= false; exit; end;

  // Add new TabSheet
  newTabSheet:= TTabSheet.Create(self);
  newTabSheet.PageControl:= FPageControl;
  FPageControl.ActivePage:= newTabSheet;

  with TApeVirtualStructureTree.Create(self) do begin
    Parent:= newTabSheet;
    Align:= alClient;
    DisplayFontName:= self.FDisplayFontName;
    Language:= FLanguage;
    ShowInfo:= FShowInfo;
    SystemId:= structureDoc.systemId;
    OnFocusChanged:= VSTFocusChanged;
    OnInfoEdited:= VSTInfoEdited;
    OnItemEdited:= VSTItemEdited;
    UpdateStructure(structureDoc);
  end;
  UpdateActiveTagCaption(GetFileDescNode(structureDoc));
end;

function TApeStructurePageControl.CanAdd: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then result:= avst.CanAdd
    else result:= false;
end;

function TApeStructurePageControl.CanDelete: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then result:= avst.CanDelete
    else result:= false;
end;

function TApeStructurePageControl.CanDown: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then result:= avst.CanDown
    else result:= false;
end;

function TApeStructurePageControl.CanEdit: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then result:= avst.CanEdit
    else result:= false;
end;

function TApeStructurePageControl.CanIndent: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then result:= avst.CanIndent
    else result:= false;
end;

function TApeStructurePageControl.CanPullOut: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then result:= avst.CanPullOut
    else result:= false;
end;

function TApeStructurePageControl.CanUp: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then result:= avst.CanUp
    else result:= false;
end;

function TApeStructurePageControl.CloseStructure(const index: integer): boolean;
var
  newActPageIndex: integer;
begin
  with FPageControl do begin
    // Out of bounds?
    if (index < 0) or (index > pred(PageCount))
      then begin result:= false; exit; end
      else result:= true;

    // Calculate new active tab
    if index > ActivePageIndex then begin
      newActPageIndex:= ActivePageIndex;
    end else if index < ActivePageIndex then begin
      newActPageIndex:= pred(ActivePageIndex);
    end else if pred(PageCount) = ActivePageIndex
      then newActPageIndex:= pred(ActivePageIndex)
      else newActPageIndex:= ActivePageIndex;

    Pages[index].Free;

    // Apply new active tab (applying it after deleting the old active tab looks nicer)
    ActivePageIndex:= newActPageIndex;

  end; {with ...}
end;

function TApeStructurePageControl.Delete: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then begin avst.Delete; result:= true; end
    else result:= false;
end;

function TApeStructurePageControl.Down: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then begin avst.Down; result:= true; end
    else result:= false;
end;

function TApeStructurePageControl.DocumentAsString(
  const index: integer): string;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(index);
  if assigned(avst)
    then result:= avst.DocumentAsString
    else result:= '';
end;

function TApeStructurePageControl.DocumentAsWideString(
  const index: integer): wideString;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(index);
  if assigned(avst)
    then result:= avst.DocumentAsWideString
    else result:= '';
end;

function TApeStructurePageControl.Indent: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then begin avst.Indent; result:= true; end
    else result:= false;
end;

function TApeStructurePageControl.InfoEdit: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then begin avst.InfoEdit; result:= true; end
    else result:= false;
end;

function TApeStructurePageControl.ItemEdit: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then begin avst.ItemEdit; result:= true; end
    else result:= false;
end;

function TApeStructurePageControl.PullOut: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then begin avst.PullOut; result:= true; end
    else result:= false;
end;

function TApeStructurePageControl.Up: boolean;
var
  avst: TApeVirtualStructureTree;
begin
  avst:= GetVirtualStructureTree(FPageControl.ActivePageIndex);
  if assigned(avst)
    then begin avst.Up; result:= true; end
    else result:= false;
end;

{ TApeStructureReader }

function TApeStructureReader.parseloopItem(const BVT: TBaseVirtualTree;
                                           const Node: PVirtualNode): boolean;
var
  attributeList: TdomNameValueList;
  Data: PTreeData;
  el: TdomElement;
  i: integer;
  loc: TdomLocator;
begin
  loc:= nil;  // Perhaps replace this with a real Locator in a future version.
  result:= false;
  attributeList:= TdomNameValueList.create;
  try
    if assigned(Node) then begin
      data:= BVT.GetNodeData(Node);
      el:= data.item;
      if assigned(el) then begin
        attributeList.clear;
        if not writeStartElement(loc,'ape:item',attributeList) then exit;

        with el.childNodes do
          for i:= 0 to pred(length) do
            if not parseloop(loc,item(i)) then exit;

        if Node.ChildCount > 0 then begin
          with attributeList do begin
            clear;
            add('ape:type','seq');
          end;
          if not writeStartElement(loc,'ape:list',attributeList) then exit;
          if not parseloopItem(BVT,Node.FirstChild) then exit;
          if not writeEndElement(loc,'ape:list') then exit;
        end;

        if not writeEndElement(loc,'ape:item') then exit;
      end else exit;
    end else exit;
  finally
    attributeList.Free;
  end;
  if assigned(Node.NextSibling)
    then if not parseloopItem(BVT,Node.NextSibling) then exit;
  result:= true;
end;

function TApeStructureReader.parse(const vst: TApeVirtualStructureTree): boolean;
const
  STRUCTURETABLE_DTD: wideString =
      '  <!ELEMENT ape:structureTable (ape:head,ape:body) >'
    + '  <!ELEMENT ape:head (ape:fileDesc) >'
    + '  <!ELEMENT ape:fileDesc (ape:titleStmt,ape:publicationStmt,ape:sourceDesc,ape:notesStmt,ape:renderingStmt) >'
    + '  <!ELEMENT ape:titleStmt (ape:title*,ape:author?) >'
    + '  <!ELEMENT ape:publicationStmt (ape:availability) >'
    + '  <!ELEMENT ape:sourceDesc (ape:p)* >'
    + '  <!ELEMENT ape:notesStmt (ape:note)* >'
    + '  <!ELEMENT ape:renderingStmt (ape:displayPriority) >'
    + '  <!ELEMENT ape:title (#PCDATA) >'
    + '  <!ELEMENT ape:author (#PCDATA) >'
    + '  <!ELEMENT ape:availability (ape:p)* >'
    + '  <!ELEMENT ape:p (#PCDATA) >'
    + '  <!ELEMENT ape:note (#PCDATA) >'
    + '  <!ELEMENT ape:displayPriority (ape:displayRole)* >'
    + '  <!ELEMENT ape:displayRole EMPTY >'

    + '  <!ELEMENT ape:body (ape:list)* >'
    + '  <!ELEMENT ape:list (ape:item)* >'
    + '  <!ELEMENT ape:item (ape:caption,ape:comment,ape:resourceMap?,ape:list?) >'
    + '  <!ELEMENT ape:caption (#PCDATA) >'
    + '  <!ELEMENT ape:comment (ape:p)* >'
    + '  <!ELEMENT ape:resourceMap (ape:resource)* >'
    + '  <!ELEMENT ape:resource EMPTY >'

    + '  <!ATTLIST ape:structureTable'
    + '            ape:version   CDATA      #FIXED "1.0.0"'
    + '            xmlns:ape     CDATA      #FIXED "http://www.philo.de/ape/test-001"'
    + '            xmlns:xlink   CDATA      #FIXED "http://www.w3.org/1999/xlink"'
    + '            xml:base      CDATA      #IMPLIED >'

    + '  <!ATTLIST ape:title'
    + '            xml:lang      CDATA      #IMPLIED'
    + '            ape:type      CDATA      #IMPLIED >'
    + '  <!ATTLIST ape:availability'
    + '            ape:status    CDATA      #IMPLIED >'
    + '  <!ATTLIST ape:p'
    + '            xml:lang      CDATA      #IMPLIED >'
    + '  <!ATTLIST ape:note'
    + '            xml:lang      CDATA      #IMPLIED'
    + '            ape:type      CDATA      #IMPLIED >'
    + '  <!ATTLIST ape:displayRole'
    + '            ape:uid       CDATA      #REQUIRED >'

    + '  <!ATTLIST ape:list'
    + '            ape:type      (seq)      #FIXED "seq" >'
    + '  <!ATTLIST ape:resource'
    + '            xlink:href    CDATA      #REQUIRED >';
var
  loc: TdomLocator;
  attributeList: TdomNameValueList;
begin
  loc:= nil;  // Perhaps replace this with a real Locator in a future version.
  result:= false;

  attributeList:= TdomNameValueList.create;
  try
    if not writeStartDocument(loc,'1.0','',STANDALONE_UNSPECIFIED) then exit;

    // DocType declaration
    if not writeDoctype(loc,'ape:structureTable','','',concat(STRUCTURETABLE_DTD)) then exit;

    // ape:structureTable
    if not writeStartPrefixMapping(loc,'ape','http://www.philo.de/ape/test-001') then exit;
    if not writeStartPrefixMapping(loc,'xlink','http://www.w3.org/1999/xlink') then exit;
    with attributeList do begin
      clear;
      add('xmlns:ape','http://www.philo.de/ape/test-001');
      add('xmlns:xlink','http://www.w3.org/1999/xlink');
    end;
    if not writeStartElement(loc,'ape:structureTable',attributeList) then exit;

    // ape:head
    attributeList.clear;
    if not writeStartElement(loc,'ape:head',attributeList) then exit;

    // ape:fileDesc
    if assigned(vst.FFileDesc) then begin
      if not parseloop(loc,vst.FFileDesc) then exit;
    end else begin
      attributeList.clear;
      if not writeStartElement(loc,'ape:fileDesc',attributeList) then exit;
      // xxx
      if not writeEndElement(loc,'ape:fileDesc') then exit;
    end;

    // ape:head finished
    if not writeEndElement(loc,'ape:head') then exit;

    // ape:body
    attributeList.clear;
    if not writeStartElement(loc,'ape:body',attributeList) then exit;

    // ape:list
    with attributeList do begin
      clear;
      add('ape:type','seq');
    end;
    if not writeStartElement(loc,'ape:list',attributeList) then exit;

    // recursively parse the items:
    if not parseloopItem(vst.FVirtualStringTree,vst.FVirtualStringTree.GetFirst) then exit;

    if not writeEndElement(loc,'ape:list') then exit;
    if not writeEndElement(loc,'ape:body') then exit;
    if not writeEndElement(loc,'ape:structureTable') then exit;
    if not writeEndDocument(loc) then exit;

  finally
    attributeList.Free;
  end;
  result:= true;
end;

end.
