unit ResourceDisplayFrame;

// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with 
// the License. You may obtain a copy of the License at 
// "http://www.mozilla.org/MPL/"
// 
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
// 
// The Original Code is "ResourceDisplayFrame.pas".
// 
// The Initial Developer of the Original Code is Dieter Khler (Heidelberg, 
// Germany, "http://www.philo.de/").
// Portions created by the Initial Developer are Copyright (C) 2002
// Dieter Khler. All Rights Reserved.
// 
// Alternatively, the contents of this file may be used under the terms of
// the GNU General Public License Version 2 or later (the "GPL"), in which 
// case the provisions of the GPL are applicable instead of those above. 
// If you wish to allow use of your version of this file only under the 
// terms of the GPL, and not to allow others to use your version of this 
// file under the terms of the MPL, indicate your decision by deleting the
// provisions above and replace them with the notice and other provisions
// required by the GPL. If you do not delete the provisions above, a
// recipient may use your version of this file under the terms of any one
// of the MPL or the GPL.

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, 
  OleCtrls, SHDocVw, StdCtrls, ComCtrls, XDOM_2_4, ToolWin, Buttons,
  ExtCtrls, VirtualTrees, ActnList, ApeUtils;

type

  TApeDisplayPacket = class(TPersistent)
  private
    FDomImplementation: TDomImplementation;
    FDoc: TdomDocument;
    FMappingList: TdomElement;
    function doAppendDisplayElement(const parentElement: TdomElement;
                                    const uid: wideString;
                                      out labelsElement,
                                          resourceListElement: TdomElement): TdomElement;
    function doAppendLabelElement(const parentElement: TdomElement;
                                  const labelText,
                                        language: wideString): TdomElement;
    function doAppendMappingElement(const uid: wideString;
                                      out labelsElement,
                                          displayListElement: TdomElement): TdomElement;
    function doAppendResourceElement(const parentElement: TdomElement;
                                     const href,
                                           editionUid: wideString): TdomElement;
    function doAppendUidElement(const parentElement: TdomElement;
                                const elementName,
                                      uidValue: wideString): TdomElement;
  public
    constructor create; virtual;
    destructor destroy; override;
    function appendDisplayLabel(const mappingUid, uid, labelText, language: wideString): boolean; virtual;
    function appendDisplay(const mappingUid, uid: wideString):boolean; virtual;
    function appendMapping(const uid: wideString): boolean; virtual;
    function appendMappingLabel(const uid, labelText, language: wideString): boolean; virtual;
    function appendResource(const mappingUid, displayUid, href, editionUid: wideString): boolean; virtual;
    function appendResourceLabel(const mappingUid, displayUid, href, labelText, language: wideString): boolean; virtual;
    procedure assign(source: TPersistent); override;
    property mappingList: TdomElement read FMappingList;
  end;

  TResDispFrame = class(TFrame)
    MappingsTabCtrl: TTabControl;
    DisplaysTabCtrl: TTabControl;
    WebBrowser1: TWebBrowser;
    DomImplementation1: TDomImplementation;
    ButtonPanel: TPanel;
    ShowBrowserBtn: TSpeedButton;
    CurtainBtn: TSpeedButton;
    VailDisplay: TVirtualStringTree;
    ResourcesTabCtrl: TTabControl;
    Panel2: TPanel;
    BrowserPanel: TPanel;
    procedure VailDisplayGetText(Sender: TBaseVirtualTree;
      Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType;
      var Text: WideString);
    procedure MappingsTabCtrlChange(Sender: TObject);
    procedure DisplaysTabCtrlChange(Sender: TObject);
    procedure ResourcesTabCtrlChange(Sender: TObject);
    procedure ShowBrowserBtnClick(Sender: TObject);
  private
    FApeDisplayPacket: TApeDisplayPacket;
    FDefaultResource: TFilename;
    FDisplayPriority: TApeDisplayPriorityList;
    FIgnorePriority: boolean;
    FLastMappingTabVersionUid: wideString;
    FLastResourceTabVersionUid: wideString;
    function GetActiveDisplay: TdomElement;
    function GetActiveMapping: TdomElement;
    function GetActiveResource: TdomElement;
  protected
    procedure SetDefaultResource(const Value: TFilename);
    procedure SetIgnorePriority(const Value: boolean); virtual;
    function UpdateBrowser(const Input: TApeDisplayPacket): boolean; virtual;
    function UpdateDisplaysTabCtrl(const DisplayList: TdomElement): boolean; virtual;
    function UpdateMappingsTabCtrl(const MappingList: TdomElement): boolean; virtual;
    function UpdateNavigator(const ResourceItem: TdomElement): boolean; virtual;
    function UpdateResourcesTabCtrl(const ResourceList: TdomElement): boolean; virtual;
    procedure UpdateVailDisplay; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Init(const Input: TApeDisplayPacket); virtual;
    procedure RefreshDisplay; virtual;
    property DisplayPriority: TApeDisplayPriorityList read FDisplayPriority;
  published
    property DefaultResource: TFilename read FDefaultResource write SetDefaultResource;
    property IgnorePriority: boolean read FIgnorePriority write SetIgnorePriority default false;
  end;

  TApeResourceDisplay = class(TResDispFrame);

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Ape', [ TApeResourceDisplay ]);
end;

{$R *.DFM}

type

TTreeData = record
  node: TdomNode;
end;

PTreeData = ^TTreeData;

{ TApeDisplayPacket }

constructor TApeDisplayPacket.create;
begin
  FDomImplementation:= TDomImplementation.create(nil);
  FDoc:= FDomImplementation.createDocument('displayPacket',nil);
  with FDoc do begin
    FMappingList:= createElement('mappingList');
    with documentElement do begin
      appendChild(FMappingList);
    end;
  end;
end;

destructor TApeDisplayPacket.destroy;
begin
  FDomImplementation.free;
  inherited;
end;

function TApeDisplayPacket.appendDisplay(const mappingUid,
                                               uid: wideString): boolean;
var
  dummy1,dummy2,displayList: TdomElement;
begin
  result:= assigned(doAppendMappingElement(mappingUid,dummy1,displayList));
  if result
    then result:= assigned(doAppendDisplayElement(displayList,uid,dummy1,dummy2));
end;

function TApeDisplayPacket.appendDisplayLabel(const mappingUid,
                                                    uid,
                                                    labelText,
                                                    language: wideString): boolean;
var
  i,j: integer;
begin
  result:= false;
  with FMappingList.childNodes do begin
    for i:= 0 to pred(length) do begin
      if TdomElement(item(i)).getAttributeNormalizedValue('uid') = mappingUid then begin
        with item(i).childnodes.item(1).childNodes do begin  // = "./mapping/displayList/"
          for j:= 0 to pred(length) do begin
            with TdomElement(item(j)) do begin
              if ( getAttributeNormalizedValue('uid') = uid )
               then result:= assigned(doAppendLabelElement(TdomElement(childnodes.item(0)),labelText,language)) or result;
            end;
          end;
        end;
      end;
    end;
  end;
end;

function TApeDisplayPacket.appendMapping(const uid: wideString): boolean;
var
  dummy1,dummy2: TdomElement;
begin
  result:= assigned(doAppendMappingElement(uid,dummy1,dummy2));
end;

function TApeDisplayPacket.appendMappingLabel(const uid,
                                                    labelText,
                                                    language: wideString): boolean;
var
  dummy,labels: TdomElement;
begin
  result:= assigned(doAppendMappingElement(uid,labels,dummy));
  if result
    then result:= assigned(doAppendLabelElement(labels,labelText,language));
end;

function TApeDisplayPacket.appendResource(const mappingUid,
                                                displayUid,
                                                href,
                                                editionUid: wideString): boolean;
var
  dummy,displayList,resourceList: TdomElement;
begin
  result:= assigned(doAppendMappingElement(mappingUid,dummy,displayList));
  if result then begin
    result:= assigned(doAppendDisplayElement(displayList,displayUid,dummy,resourceList));
    if result
      then result:= assigned(doAppendResourceElement(resourceList,href,editionUid));
  end;
end;

function TApeDisplayPacket.appendResourceLabel(const mappingUid,
                                                     displayUid,
                                                     href,
                                                     labelText,
                                                     language: wideString): boolean;
var
  i,j,k: integer;
begin
  result:= false;
  with FMappingList.childNodes do begin
    for i:= 0 to pred(length) do begin
      if TdomElement(item(i)).getAttributeNormalizedValue('uid') = mappingUid then begin
        with item(i).childnodes.item(1).childNodes do begin  // = "./mapping/displayList/"
          for j:= 0 to pred(length) do begin
            if TdomElement(item(j)).getAttributeNormalizedValue('uid') = displayUid then begin
              with item(j).childnodes.item(1).childNodes do begin  // = "./mapping/resourceList/"
                for k:= 0 to pred(length) do begin
                  with TdomElement(item(k)) do begin
                    if getAttributeNormalizedValue('href') = href
                     then result:= assigned(doAppendLabelElement(TdomElement(childnodes.item(0)),labelText,language)) or result;
                  end;
                end;
              end;
            end;
          end;
        end;
      end;
    end;
  end;
end;

procedure TApeDisplayPacket.assign(source: TPersistent);
begin
  if source is TApeDisplayPacket then begin
    if source = self then exit;
    with FDoc do begin
      FMappingList:= importNode(TApeDisplayPacket(source).FMappingList,true) as TdomElement;
      with documentElement do begin
        clear;
        appendChild(FMappingList);
      end;
    end;
    exit;
  end;
  inherited;
end;

function TApeDisplayPacket.doAppendDisplayElement(const parentElement: TdomElement;
                                                  const uid: wideString;
                                                    out labelsElement,
                                                        resourceListElement: TdomElement): TdomElement;
var
  index: integer;
  item_el: TdomElement;
begin
  result:= nil;
  if uid <> '' then begin

    // Find out whether the Display to append already exists?
    with parentElement.childNodes do begin
      for index:= 0 to pred(length) do begin
        item_el:= TdomElement(item(index));
        if (item_el.getAttributeNormalizedValue('uid') = uid) then begin
          result:= item_el;
          labelsElement:= TdomElement(result.childNodes.item(0));
          resourceListElement:= TdomElement(result.childNodes.item(1));
          exit;
        end;
      end;
    end;

    // If the Display to append does not already exists, then generate it:
    result:= FDoc.createElement('display');
    result.setAttribute('uid',uid);
    labelsElement:= FDoc.createElement('labels');
    resourceListElement:= FDoc.createElement('resourceList');
    result.appendChild(labelsElement);
    result.appendChild(resourceListElement);
    parentElement.appendChild(result);

  end else begin
    labelsElement:= nil;
    resourceListElement:=  nil;
  end;
end;

function TApeDisplayPacket.doAppendLabelElement(const parentElement: TdomElement;
                                                const labelText,
                                                      language: wideString): TdomElement;
var
  index: integer;
  item_el: TdomElement;
begin
  result:= nil;
  if labelText = '' then exit;

  // Find out whether the Label to append already exists?
  with parentElement.childNodes do begin
    for index:= 0 to pred(length) do begin
      item_el:= TdomElement(item(index));
      if (item_el.getAttributeNormalizedValue('lang') = language) and (item_el.textContent = labelText) then begin
        result:= item_el;
        exit;
      end;
    end;
  end;

  // If the Label to append does not already exists, then generate it:
  result:= FDoc.createElement('label');
  if language <> ''
    then result.setAttribute('lang',language);
  result.appendChild(FDoc.createTextNode(labelText));
  parentElement.appendChild(result);
end;

function TApeDisplayPacket.doAppendMappingElement(const uid: wideString;
                                                    out labelsElement,
                                                        displayListElement: TdomElement): TdomElement;
begin
  result:= doAppendUidElement(FMappingList,'mapping',uid);
  if assigned(result) then begin
     if not result.hasChildNodes then begin
       labelsElement:= FDoc.createElement('labels');
       result.appendChild(labelsElement);
       displayListElement:= FDoc.createElement('displayList');
       result.appendChild(displayListElement);
     end else begin
       labelsElement:= TdomElement(result.childNodes.item(0));
       displayListElement:= TdomElement(result.childNodes.item(1));
     end;;
  end else begin
    labelsElement:= nil;
    displayListElement:=  nil;
  end;
end;

function TApeDisplayPacket.doAppendResourceElement(const parentElement: TdomElement;
                                                   const href,
                                                         editionUid: wideString): TdomElement;
var
  index: integer;
  item_el: TdomElement;
begin
  result:= nil;
  if href = '' then exit;

  // Find out whether the Resource to append already exists?
  with parentElement.childNodes do begin
    for index:= 0 to pred(length) do begin
      item_el:= TdomElement(item(index));
      if (item_el.getAttributeNormalizedValue('href') = href) then begin
        result:= item_el;
        exit;
      end;
    end;
  end;

  // If the Resource to append does not already exists, then generate it:
  result:= FDoc.createElement('resource');
  result.setAttribute('href',href);
  result.setAttribute('uid',editionUid);
  result.appendChild(FDoc.createElement('labels'));
  parentElement.appendChild(result);
end;

function TApeDisplayPacket.doAppendUidElement(const parentElement: TdomElement;
                                              const elementName,
                                                    uidValue: wideString): TdomElement;
var
  index: integer;
  item_el: TdomElement;
begin
  result:= nil;
  if (uidValue = '') or (elementName = '') or not assigned(parentElement) then exit;

  // Find out whether the Element to append already exists?
  with parentElement.childNodes do begin
    for index:= 0 to pred(length) do begin
      item_el:= TdomElement(item(index));
      if (item_el.nodeName = elementName) and (item_el.getAttributeNormalizedValue('uid') = uidValue) then begin
        result:= item_el;
        exit;
      end;
    end;
  end;

  // If the Element to append does not already exists, then generate it:
  result:= FDoc.createElement(elementName);
  result.setAttribute('uid',uidValue);
  parentElement.appendChild(result);
end;

{ TResDispFrame }

constructor TResDispFrame.Create(AOwner: TComponent);
begin
  inherited;
  FIgnorePriority:= false;
  FDisplayPriority:= TApeDisplayPriorityList.Create;
end;

destructor TResDispFrame.Destroy;
begin
  if assigned(FApeDisplayPacket)
    then FApeDisplayPacket.free;
  FDisplayPriority.free;
  inherited;
end;

procedure TResDispFrame.Init(const Input: TApeDisplayPacket);
begin
  if assigned(FApeDisplayPacket)
    then FApeDisplayPacket.free;
  FApeDisplayPacket:= Input;
  UpdateBrowser(Input);
  UpdateVailDisplay;
end;

function TResDispFrame.GetActiveDisplay: TdomElement;
begin
  with DisplaysTabCtrl do
    if TabIndex > -1 then begin
      if Tabs.objects[TabIndex] is TdomElement
        then result:= TdomElement(Tabs.objects[TabIndex])
        else result:= nil;
    end else result:= nil;
end;

function TResDispFrame.GetActiveMapping: TdomElement;
begin
  with MappingsTabCtrl do
    if TabIndex > -1 then begin
      if Tabs.objects[TabIndex] is TdomElement
        then result:= TdomElement(Tabs.objects[TabIndex])
        else result:= nil;
    end else result:= nil;
end;

function TResDispFrame.GetActiveResource: TdomElement;
begin
  with ResourcesTabCtrl do
    if TabIndex > -1 then begin
      if Tabs.objects[TabIndex] is TdomElement
        then result:= TdomElement(Tabs.objects[TabIndex])
        else result:= nil;
    end else result:= nil;
end;

function TResDispFrame.UpdateBrowser(const Input: TApeDisplayPacket): boolean;
begin
  if assigned(Input)
    then result:= UpdateMappingsTabCtrl(Input.MappingList)
    else result:= UpdateMappingsTabCtrl(nil);
end;

function TResDispFrame.UpdateDisplaysTabCtrl(const DisplayList: TdomElement): boolean;
var
  activeDisplayEl,displayEl,labelEl,resourceListEl: TdomElement;
begin
  with DisplaysTabCtrl do begin

    // Update
    Tabs.clear;
    if assigned(DisplayList) then begin
      with DisplayList do begin
        displayEl:= getFirstChildElement('display');
        while assigned(displayEl) do begin
          labelEl:= displayEl.getFirstChildElement('labels').getFirstChildElement('label');
          if assigned(labelEl)
            then Tabs.AddObject(labelEl.textContent,displayEl);
          displayEl:= displayEl.getNextSiblingElement('display');
        end; {while ...}
      end; {with ...}
    end; {if ...}

    activeDisplayEl:= GetActiveDisplay;
    if assigned(activeDisplayEl)
      then resourceListEl:= activeDisplayEl.getFirstChildElement('resourceList')
      else resourceListEl:= nil;
    result:= UpdateResourcesTabCtrl(resourceListEl);

  end; {with VersionsTabCtrl ...}
end;

function TResDispFrame.UpdateMappingsTabCtrl(const MappingList: TdomElement): boolean;
var
  i: integer;
  activeMappingEl,displayListEl,mappingEl,labelEl: TdomElement;
  mappingFound: boolean;
begin
  with MappingsTabCtrl do begin

    Tabs.BeginUpdate;

    try
      // Update
      Tabs.clear;
      if assigned(MappingList) then begin
        with MappingList do begin

          mappingFound:= false;
          if not IgnorePriority then begin
            for i:= 0 to pred(DisplayPriority.count) do begin
              mappingEl:= getFirstChildElement('mapping');
              while assigned(mappingEl) do begin
                if mappingEl.getAttributeNormalizedValue('uid') = DisplayPriority[i] then begin
                  labelEl:= mappingEl.getFirstChildElement('labels').getFirstChildElement('label');
                  if assigned(labelEl)
                    then Tabs.AddObject(labelEl.textContent,mappingEl);
                  mappingFound:= true;
                  break;
                end;
                mappingEl:= mappingEl.getNextSiblingElement('mapping');
              end; {while ...}
              if mappingFound then break;
            end; {for ...}
          end;

          if not mappingFound then begin
            mappingEl:= getFirstChildElement('mapping');
            while assigned(mappingEl) do begin
              labelEl:= mappingEl.getFirstChildElement('labels').getFirstChildElement('label');
              if assigned(labelEl)
                then Tabs.AddObject(labelEl.textContent,mappingEl);
              mappingEl:= mappingEl.getNextSiblingElement('mapping');
            end; {while ...}
          end;

       end;
     end;

      // Restore the type of displayed document version
      for i:= 0 to pred(Tabs.Count) do
        if TdomElement(Tabs.objects[i]).getAttributeNormalizedValue('uid') = FLastMappingTabVersionUid
          then begin TabIndex:= i; break; end;

      activeMappingEl:= GetActiveMapping;
      if assigned(activeMappingEl)
        then displayListEl:= activeMappingEl.getFirstChildElement('displayList')
        else displayListEl:= nil;
      result:= UpdateDisplaysTabCtrl(displayListEl);

    finally
      Tabs.EndUpdate;
    end;
  end; {with VersionsTabCtrl ...}
end;

function TResDispFrame.UpdateResourcesTabCtrl(const ResourceList: TdomElement): boolean;
var
  i: integer;
  activeResourceEl,resourceEl,labelEl: TdomElement;
begin
  with ResourcesTabCtrl do begin

    Tabs.BeginUpdate;

    try
      // Update
      Tabs.clear;
      if assigned(ResourceList) then begin
        with ResourceList do begin
          resourceEl:= getFirstChildElement('resource');
          while assigned(resourceEl) do begin
            labelEl:= resourceEl.getFirstChildElement('labels').getFirstChildElement('label');
            if assigned(labelEl)
              then Tabs.AddObject(labelEl.textContent,resourceEl);
            resourceEl:= resourceEl.getNextSiblingElement('resource');
          end; {while ...}
        end; {with ...}
      end; {if ...}

      // Restore the type of displayed document version
      for i:= 0 to pred(Tabs.Count) do
        if TdomElement(Tabs.objects[i]).getAttributeNormalizedValue('uid') = FLastResourceTabVersionUid
          then begin TabIndex:= i; break; end;

      activeResourceEl:= GetActiveResource;
      result:= UpdateNavigator(activeResourceEl)

    finally
      Tabs.EndUpdate;
    end;
  end; {with VersionsTabCtrl ...}
end;

function TResDispFrame.UpdateNavigator(const ResourceItem: TdomElement): boolean;
begin
  if assigned(ResourceItem) then begin
    WebBrowser1.Navigate(ResourceItem.getAttributeNormalizedValue('href'));
    result:= true;
  end else begin
    if DefaultResource <> ''
      then WebBrowser1.Navigate(DefaultResource);
    result:= false;
  end;
end;

procedure TResDispFrame.UpdateVailDisplay;

  procedure HandleNodeList(pn: PVirtualNode; DomNodeList:TDomNodeList);
  var
    i:integer;
    DomNode:TDomNode;
    tn: PVirtualNode;
    Data: PTreeData;
  begin
    for i:=0 to pred(DomNodeList.length) do begin
      tn:= VailDisplay.AddChild(pn);
      Data:= VailDisplay.GetNodeData(tn);
      DomNode:= DomNodeList.Item(i);
      data.node:= DomNode;
      if assigned(DomNode.ChildNodes)
        then HandleNodeList(tn, DomNode.ChildNodes);
    end;
  end;

begin
  with VailDisplay do begin
    Clear;
    NodeDataSize := SizeOf(TTreeData);
    if assigned(FApeDisplayPacket)
      then HandleNodeList(nil,FApeDisplayPacket.FDoc.ChildNodes);
    FullExpand;
  end; {with ...}
end;

procedure TResDispFrame.DisplaysTabCtrlChange(Sender: TObject);
var
  activeDisplayEl: TdomElement;
begin
  DisplaysTabCtrl.Tabs.BeginUpdate;
  try
    activeDisplayEl:= GetActiveDisplay;
    if assigned(activeDisplayEl)
      then UpdateResourcesTabCtrl(activeDisplayEl.getFirstChildElement('resourceList'));
  finally
    DisplaysTabCtrl.Tabs.EndUpdate;
  end;
end;

procedure TResDispFrame.MappingsTabCtrlChange(Sender: TObject);
var
  activeMappingEl: TdomElement;
begin
  MappingsTabCtrl.Tabs.BeginUpdate;
  try
    activeMappingEl:= GetActiveMapping;
    if assigned(activeMappingEl) then begin
      UpdateDisplaysTabCtrl(activeMappingEl.getFirstChildElement('displayList'));
      FLastMappingTabVersionUid:= activeMappingEl.getAttributeNormalizedValue('uid')
    end else FLastMappingTabVersionUid:= '';
  finally
    MappingsTabCtrl.Tabs.EndUpdate;
  end;
end;

procedure TResDispFrame.RefreshDisplay;
begin
  UpdateBrowser(FApeDisplayPacket);
end;

procedure TResDispFrame.ResourcesTabCtrlChange(Sender: TObject);
var
  activeResourceEl: TdomElement;
begin
  ResourcesTabCtrl.Tabs.BeginUpdate;
  try
    activeResourceEl:= GetActiveResource;
    if assigned(activeResourceEl) then begin
      UpdateNavigator(activeResourceEl);
      FLastResourceTabVersionUid:= activeResourceEl.getAttributeNormalizedValue('uid')
    end else FLastResourceTabVersionUid:= '';
  finally
    ResourcesTabCtrl.Tabs.EndUpdate;
  end;
end;

procedure TResDispFrame.VailDisplayGetText(Sender: TBaseVirtualTree;
  Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType;
  var Text: WideString);
var
  domNode: TdomNode;
  Data: PTreeData;
  i: integer;
begin
  data:= Sender.GetNodeData(Node);
  domNode:= data.node;
  sender.Font.Color:= clBlack;
  if assigned(domNode) then with domNode do begin
    case nodeType of
      ntElement_Node: begin
        text:= '<' + NodeName;
        for i:= 0 to pred(attributes.length) do begin
          text:= text+' '+attributes.item(i).nodeName+'="'+attributes.item(i).nodeValue+'"';
          if NodeValue <> '' then text:= text + ' | ' + NodeValue;
        end;
        text:= text + '>';
      end;
      ntText_Node: begin
        text:= nodeValue;
      end;
    else begin
        text:= '[' + NodeName + '] ';
        if NodeValue <> '' then text:= text + NodeValue;
      end;
    end; {case ...}
  end else text:= '';
end;

procedure TResDispFrame.ShowBrowserBtnClick(Sender: TObject);
begin
  if Sender = ShowBrowserBtn
    then BrowserPanel.Show;
  if Sender = CurtainBtn
    then BrowserPanel.Hide;
end;

procedure TResDispFrame.SetDefaultResource(const Value: TFilename);
begin
  FDefaultResource := Value;
  RefreshDisplay;
end;

procedure TResDispFrame.SetIgnorePriority(const Value: boolean);
begin
  FIgnorePriority := Value;
end;

end.
